# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Line shapes for peak fitting.
"""

from typing import Optional
import numpy as np


class Peak:
    def __init__(self, position: float, width: float, area: float) -> None:
        """
        Base class for a peak with position, width, and intensity.

        Parameters:
        - position (float): Position of the peak on the energy axis.
        - width (float): Full width at half maximum (FWHM) of the peak.
        - area (float): Area of the peak.
        """
        if width <= 0:
            raise ValueError("Width must be positive.")
        if area <= 0:
            raise ValueError("Intensity must be positive.")
        self.position = position
        self.width = width
        self.area = area


class Gaussian(Peak):
    """
    Gaussian peak with specified position, width, and area.
    """

    def calc_lineshape(self, x: float) -> Optional[float]:
        """
        Calculate the Gaussian lineshape at a given energy value x.

        Parameters:
        - x (float): Energy value on the energy axis.

        Returns:
        - float: Value of the Gaussian lineshape at x, or None if width is zero.
        """
        if self.width != 0:
            # Calculate intensity from area
            intensity = self.area / (np.sqrt(2 * np.pi) * self.width)
            exponent = -4 * np.log(2) * (x - self.position) ** 2 / self.width**2
            return intensity * np.exp(exponent)
        return None

    def formula(self) -> str:
        """
        Returns a string representation of the Gaussian formula.

        Returns:
        - str: The formula used for the Gaussian lineshape.
        """
        return "G(x; position, width, area) = (area / (sqrt(2 * pi) * width)) * exp[-(4 * ln(2) * (x - position)^2) / width^2]"

    def __repr__(self) -> str:
        return (
            f"Gaussian(position={self.position}, width={self.width}, area={self.area})"
        )


class Lorentzian(Peak):
    """Lorentzian peak with specified position, width, and area."""

    def calc_lineshape(self, x: float) -> Optional[float]:
        """
        Calculate the Lorentzian lineshape at a given energy value x.

        Parameters:
        - x (float): Energy value on the energy axis.

        Returns:
        - float: Value of the Lorentzian lineshape at x, or None if width is zero.
        """
        if self.width != 0:
            # Calculate intensity from area
            intensity = self.area / (np.pi * self.width)
            return intensity / (1 + (4 * (x - self.position) ** 2) / self.width**2)
        return None

    def formula(self) -> str:
        """
        Returns a string representation of the Lorentzian formula.

        Returns:
        - str: The formula used for the Lorentzian lineshape.
        """
        return "L(x; position, width, area) = (area / (pi * width)) / (1 + (4 * (x - position)^2) / width^2)"

    def __repr__(self) -> str:
        return f"Lorentzian(position={self.position}, width={self.width}, area={self.area})"


class LorentzianAsymmetric:
    def __init__(
        self, position: float, width: float, area: float, alpha: float, beta: float
    ):
        """
        Initialize the LorentzianAsymmetric profile with the required parameters.

        Parameters:
            position (float): The peak position (center).
            width (float): Full width at half maximum (FWHM) of the peak.
            area (float): Area under the peak.
            alpha (float): Scaling factor for width on the right.
            beta (float): Scaling factor for width on the left.
        """
        self.position = position
        self.width = width
        self.area = area
        self.alpha = alpha
        self.beta = beta

    def calc_lineshape(self, x: np.ndarray) -> np.ndarray:
        """
        Calculate the asymmetric Lorentzian lineshape for an array of energy values x.

        Parameters:
        - x (np.ndarray): Array of energy values on the energy axis.

        Returns:
        - np.ndarray: Array of values of the Lorentzian lineshape for each x.
        """
        # Calculate intensity from area
        intensity = self.area / (np.pi * self.width)

        # Calculate the width based on the position and scaling factors alpha and beta
        width = np.where(
            x < self.position, self.width / self.alpha, self.width / self.beta
        )

        # Calculate the Lorentzian lineshape for each value in x
        return intensity / (1 + (4 * (x - self.position) ** 2) / width**2)

    def formula(self) -> str:
        """
        Returns a string representation of the asymmetric Lorentzian formula.

        Returns:
        - str: The formula used for the asymmetric Lorentzian lineshape.
        """
        return (
            "LA(x; position, width, area, alpha, beta) = (area / (pi * width)) / (1 + ((x - position) / (width / alpha))^2) for x <= position, "
            "(area / (pi * width)) / (1 + ((x - position) / (width / beta))^2) for x > position"
        )

    def __repr__(self) -> str:
        """
        Return a string representation of the LorentzianAsymmetric object.
        """
        return (
            f"LorentzianAsymmetric(position={self.position}, width={self.width}, "
            f"area={self.area}, alpha={self.alpha}, beta={self.beta})"
        )


class LorentzianFinite(LorentzianAsymmetric):
    """Finite Lorentzian peak with specified position, width, area, asymmetry parameters, and damping."""

    def __init__(
        self,
        position: float,
        width: float,
        area: float,
        alpha: float,
        beta: float,
        w: float,
        gauss_contribution: float,
        no_of_convolutions: int,
    ):
        super().__init__(position, width, area, alpha, beta)
        self.w = w
        self.gauss_contribution = gauss_contribution
        self.no_of_convolutions = no_of_convolutions

    def calc_lineshape(self, x: np.ndarray) -> Optional[np.ndarray]:
        """
        Calculate the finite Lorentzian lineshape with damping for an array of energy values x.

        Parameters:
        - x (np.ndarray): Array of energy values on the energy axis.

        Returns:
        - np.ndarray: Array of finite Lorentzian lineshape values for each x.
        """
        # Get the Lorentzian lineshape for the array of x
        lorentzian = super().calc_lineshape(x)

        # If lorentzian is not None, apply damping
        if lorentzian is not None:
            damping_factor = 1 / (1 + 4 * ((x - self.position) / self.w) ** 2)
            return lorentzian * damping_factor

        # Return None if lorentzian is None
        return None

    def formula(self) -> str:
        """
        Returns a string representation of the finite Lorentzian formula with damping.

        Returns:
        - str: The formula used for the finite Lorentzian lineshape.
        """
        return (
            "LF(x; position, width, w) = (area / (pi * width)) / (1 + (4 * (x - position)^2) / width^2) * "
            "(1 / (1 + 4 * ((x - position) / w)^2))"
        )

    def __repr__(self) -> str:
        """
        Return a string representation of the LorentzianFinite object.
        """
        return (
            f"LorentzianFinite(position={self.position}, width={self.width}, "
            f"area={self.area}, alpha={self.alpha}, beta={self.beta}, w={self.w})"
        )


class GaussianLorentzianSum(Peak):
    """Combined Gaussian and Lorentzian peak using the existing Gaussian and Lorentzian classes."""

    def __init__(
        self, position: float, width: float, area: float, fraction_gauss: float = 0.5
    ) -> None:
        """
        Combined Gaussian and Lorentzian sum peak.

        Parameters:
        - position (float): Position of the peak.
        - width (float): Width of the peak.
        - area (float): Area of the peak (instead of intensity).
        - fraction_gauss (float): Fraction of the Gaussian contribution (between 0 and 1).
        """
        super().__init__(position, width, area)

        if fraction_gauss > 1.0:
            fraction_gauss /= 100
        self.fraction_gauss = fraction_gauss

    def calc_lineshape(self, x: float) -> Optional[float]:
        """
        Calculate the combined lineshape of the Gaussian and Lorentzian at x.

        Parameters:
        - x (float): Energy value on the energy axis.

        Returns:
        - float: Combined lineshape value at x, or None if width is zero.
        """
        if self.width != 0:
            # Calculate intensity from area for Gaussian and Lorentzian
            intensity = self.area / (np.pi * self.width)

            # Gaussian part (1 - fraction_gauss) contribution
            gauss_part = (1 - self.fraction_gauss) * Gaussian(
                self.position, self.width, intensity
            ).calc_lineshape(x)
            # Lorentzian part (fraction_gauss) contribution
            lorentz_part = self.fraction_gauss * Lorentzian(
                self.position, self.width, intensity
            ).calc_lineshape(x)
            return gauss_part + lorentz_part
        return None

    def formula(self) -> str:
        """
        Returns a detailed string representation of the combined Gaussian-Lorentzian formula.

        Returns:
        - str: The formula used for the combined lineshape.
        """
        # Using the formula for both Gaussian and Lorentzian
        gauss_formula = (
            "(area / (pi * width)) * exp[-(4 * ln(2) * (x - position)^2) / width^2]"
        )
        lorentz_formula = "area / (pi * width) / (1 + (4 * (x - position)^2) / width^2)"
        return (
            f"SGL(x; position, width, area, fraction_gauss): fraction_gauss * G(x) + (1 - fraction_gauss) * L(x) =  fraction_gauss * ({gauss_formula}) + "
            f"(1 - fraction_gauss) * ({lorentz_formula})"
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the GaussianLorentzianSum object, including details
        for position, width, area, and fraction_gauss.

        Returns:
        - str: The string representation of the object.
        """
        return (
            f"GaussianLorentzianSum(position={self.position}, width={self.width}, "
            f"area={self.area}, fraction_gauss={self.fraction_gauss})"
        )


class GaussianLorentzianProduct(Peak):
    def __init__(
        self, position: float, width: float, area: float, fraction_gauss: float = 0.5
    ) -> None:
        """
        Combined Gaussian and Lorentzian product peak.

        Parameters:
        - position (float): Position of the peak.
        - width (float): Width of the peak.
        - area (float): Area of the peak (instead of intensity).
        - fraction_gauss (float): Fraction of the Gaussian contribution.
        """
        super().__init__(position, width, area)

        if fraction_gauss > 1.0:
            fraction_gauss /= 100
        self.fraction_gauss = fraction_gauss

    def calc_lineshape(self, x: float) -> Optional[float]:
        """
        Calculate the combined lineshape of the Gaussian and Lorentzian product at x.

        Parameters:
        - x (float): Energy value on the energy axis.

        Returns:
        - float: Combined lineshape value at x, or None if width is zero.
        """
        if self.width != 0:
            # Calculate intensity from area for Gaussian and Lorentzian
            intensity = self.area / (np.pi * self.width)

            # Gaussian part (1 - fraction_gauss) contribution
            gauss_part = self.fraction_gauss * Gaussian(
                self.position, self.width, intensity
            ).calc_lineshape(x)
            # Lorentzian part (fraction_gauss) contribution
            lorentz_part = (1 - self.fraction_gauss) * Lorentzian(
                self.position, self.width, intensity
            ).calc_lineshape(x)
            return gauss_part * lorentz_part
        return None

    def formula(self) -> str:
        """
        Returns a detailed string representation of the combined Gaussian-Lorentzian formula.

        Returns:
        - str: The formula used for the combined lineshape.
        """
        # Using the formula for both Gaussian and Lorentzian
        gauss_formula = (
            "(area / (pi * width)) * exp[-(4 * ln(2) * (x - position)^2) / width^2]"
        )
        lorentz_formula = "area / (pi * width) / (1 + (4 * (x - position)^2) / width^2)"
        return (
            f"GL(x; position, width, area, fraction_gauss): fraction_gauss * G(x) * (1 - fraction_gauss) * L(x) = fraction_gauss * ({gauss_formula}) * "
            f"(1 - fraction_gauss) * ({lorentz_formula})"
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the GaussianLorentzianProduct object, including details
        for position, width, area, and fraction_gauss.

        Returns:
        - str: The string representation of the object.
        """
        return (
            f"GaussianLorentzianProduct(position={self.position}, width={self.width}, "
            f"area={self.area}, fraction_gauss={self.fraction_gauss})"
        )


class DoniachSunjic(Peak):
    """Doniach-Sunjic profile for XPS peaks with asymmetry."""

    def __init__(self, position: float, width: float, area: float, beta: float) -> None:
        """
        Initialize the Doniach-Sunjic profile with the required parameters.

        Parameters:
            position (float): The peak position (center).
            width (float): Full width at half maximum (FWHM) of the peak.
            area (float): Area under the peak (instead of intensity).
            beta (float): Asymmetry parameter (1 for symmetric Lorentzian, <1 for left skew, >1 for right skew).
        """
        # Initialize the parent Peak class with area
        super().__init__(position, width, area)
        self.beta = beta

    def calc_lineshape(self, x: np.ndarray) -> np.ndarray:
        """
        Calculate the Doniach-Sunjic profile at each energy point.

        Parameters:
            x (np.ndarray or float): The energy values at which to evaluate the profile.

        Returns:
            np.ndarray or float: The intensity values at each energy point.
        """
        # Calculate the intensity from area for normalization
        intensity = self.area / (np.pi * self.width)

        # Calculate the Doniach-Sunjic profile
        return intensity / ((1 + ((x - self.position) / self.width) ** 2) ** self.beta)

    def formula(self) -> str:
        """
        Returns the formula used for the Doniach-Sunjic profile.

        Returns:
            str: The formula for the Doniach-Sunjic profile.
        """
        return (
            "Doniach-Sunjic Profile Formula:\n"
            "f(x; area, position, width, beta, fraction_gauss)) = area / (pi * width) * ((1 + ((x - position) / width)^2)^beta)\n"
            "Where:\n"
            "  area: Area under the peak\n"
            "  position: Peak position (center)\n"
            "  width: FWHM (full width at half maximum)\n"
            "  beta: Asymmetry parameter (1 for symmetric Lorentzian)\n"
            "  x: Energy values\n"
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the DoniachSunjic object, including details
        for position, width, area, and beta.

        Returns:
        - str: The string representation of the object.
        """
        return (
            f"DoniachSunjic(position={self.position}, width={self.width}, "
            f"area={self.area}, beta={self.beta})"
        )



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Data model for CasaXPS.
"""

import re
import logging
from typing import List, Dict, Any, Optional
from dataclasses import dataclass, field
import numpy as np

from pynxtools_xps.reader_utils import XpsDataclass
from pynxtools_xps.value_mappers import convert_energy_type

from pynxtools_xps.models.lineshapes import (
    Peak,
    LorentzianAsymmetric,
    LorentzianFinite,
    GaussianLorentzianSum,
    GaussianLorentzianProduct,
    DoniachSunjic,
)

from pynxtools_xps.models.backgrounds import (
    LinearBackground,
    Shirley,
    StepUp,
    StepDown,
    TougaardU3,
    TougaardU4,
)


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def split_after_letters(string: str):
    """
    Splits the part of the string after the leading letters by commas.

    Parameters:
        string (str): The input string.

    Returns:
        tuple: A tuple with the leading letters and the list of values after splitting by commas.
    """
    # Extract the leading letters
    match = re.match(r"([A-Za-z\s]+)", string)
    leading_letters = match.group(0) if match else ""

    # Remove the leading letters and split the remaining part by commas
    remaining_part = string[len(leading_letters) :]
    split_values = remaining_part.split(",")

    if split_values == [""]:
        return leading_letters, []

    return leading_letters, [float(val) for val in split_values]


class CasaProcess:
    """Processing and fitting information from CasaXPS."""

    def __init__(self):
        self.casa_data: Dict[str, List[Any]] = {
            "energy_calibrations": [],
            "intensity_calibrations": [],
            "smoothings": [],
            "regions": [],
            "components": [],
        }

    def process_comments(self, comment_list: List[str]):
        line_map = {
            "Calib": self.process_energy_calibration,
            "KECalib": self.process_intensity_calibration,
            "TransmissionSet": self.process_intensity_calibration,
            "InCalib": self.process_intensity_calibration,
            "Sm": self.process_smoothing,
            "CASA region": self.process_region,
            "CASA comp": self.process_component,
        }

        n_processes, n_annotations, n_regions, n_components = [
            int(element)
            for element in comment_list
            if isinstance(element, str) and element.isdigit()
        ][:4]
        n_annotations *= 2  # account for empty line

        self.no_of_casa_lines = (
            5 + n_processes + n_annotations + n_regions + n_components
        )

        for line in comment_list[: self.no_of_casa_lines]:
            for token, func in line_map.items():
                if line.startswith(token):
                    func(line)

        return self.casa_data

    def flatten_metadata(self):
        # Write process data
        process_keys: List[str] = [
            "energy_calibrations",
            "intensity_calibrations",
            "smoothings",
            "regions",
            "components",
        ]

        flattened_dict: Dict[str, Any] = {}

        for process_key in process_keys:
            processes = self.casa_data[process_key]
            for i, process in enumerate(processes):
                spectrum_key = f"{process_key.rstrip('s')}{i}"
                for key, value in process.dict().items():
                    key = key.replace("_units", "/@units")
                    flattened_dict[f"{spectrum_key}/{key}"] = value

        return flattened_dict

    def process_energy_calibration(self, calib_str: str):
        """
        Process one alignment.

        Parameters
        ----------
        align_str : str
            String with alignment definition in a VMS file.
            Typical example:
                'Calib M = 6.8 A = 0 BE ADD'


        Returns
        -------
        None.

        """
        split_line = self._split_line(calib_str)

        energy_calib = CasaEnergyCalibration()

        measured_index = split_line.index("M")
        aligned_index = split_line.index("A")
        try:
            energy_calib.measured_energies = [float(split_line[measured_index + 2])]
        except ValueError:
            energies = [
                float(energy) for energy in split_line[measured_index + 2].split(",")
            ]
            energy_calib.measured_energies = energies
            energy_calib.range_calibration = True
        energy_calib.aligned_energy = float(split_line[aligned_index + 2])

        energy_calib.energy_type = convert_energy_type(split_line[7])
        energy_calib.operation = split_line[8]

        if energy_calib.operation == "ADD":
            energy_calib.energy_offset = (
                min(energy_calib.measured_energies) - energy_calib.aligned_energy
            )
            energy_calib.operation = "addition"

        energy_calib.validate_types()

        self.casa_data["energy_calibrations"] += [energy_calib]

    def process_intensity_calibration(self, calib_str: str):
        """Process different intensity calibrations."""

        intensity_calib = CasaIntensityCalibration()

        def get_power_law_coefficient(line: str):
            intensity_calib.power_law_coefficient = float(line.split("KECalib ")[1])

        def get_transmission_function_ordinate(line: str):
            intensity_calib.tf_function_ordinate_no = int(
                line.split("TransmissionSet 1.0 ")[1]
            )

        def get_both_calibrations(line: str):
            split_line = line.split(" ")
            intensity_calib.power_law_coefficient = float(split_line[1])
            intensity_calib.tf_function_ordinate_no = int(split_line[2])

        calib_map = {
            "KECalib": get_power_law_coefficient,
            "TransmissionSet": get_transmission_function_ordinate,
            "InCalib": get_both_calibrations,
        }

        for token, func in calib_map.items():
            if calib_str.startswith(token):
                func(calib_str)

        intensity_calib.validate_types()

        self.casa_data["intensity_calibrations"] += [intensity_calib]

    def process_smoothing(self, region_str: str):
        # TODO: parse smoothing of data
        pass

    def process_region(self, region_str: str):
        """
        Process one fit region.

        Parameters
        ----------
        region_str : str
            String with region definition in a VMS file.
            Typical example:
                'CASA region (*O1s*) (*Shirley*) 718.88221 726.6 9.74 2 0 9 1450.5943 -450 0 0 (*Mo 3d*) 95.9219 0 9.74'

        Returns
        -------
        None.

        """
        split_line = self._split_line(region_str)

        region = CasaRegion()
        region.name = split_line[2]
        region.bg_type = split_line[3]
        region.start = float(split_line[4])
        region.end = float(split_line[5])
        region.rsf = float(split_line[6])
        region.av_width = float(split_line[7])
        region.start_offset = float(split_line[8])
        region.end_offset = float(split_line[9])
        region.cross_section = split_line[10:14]
        region.tag = split_line[14]
        region.unknown_0 = float(split_line[15])
        region.unknown_1 = float(split_line[16])
        region.rsf_effective = float(split_line[17])

        region.validate_types()

        self.casa_data["regions"] += [region]

    def process_component(self, component_str: str):
        """
        Process one fit component if it is defined in the

        Parameters
        ----------
        component_str : str
            String with component definition in a VMS file.
            Typical example:
                'CASA comp (*O1s, lattice*) (*GL(50)*) Area 308996.39 1e-020 15401311 -1 1 MFWHM 1.5315216 0.32 2 -1 1 Position 723.30224 719.3797 726.3 -1 1 RSF 0.063 MASS 15.9994 INDEX -1 (*O 1s*) CONST (**) UNCORRECTEDRSF 2.85'


        Returns
        -------
        None.

        """
        split_line = self._split_line(component_str)

        component = CasaComponent()
        component.name = split_line[2]
        component.lineshape = split_line[3]

        area_index = split_line.index("Area")
        component.area = float(split_line[area_index + 1])
        component.area_min = float(split_line[area_index + 2])
        component.area_max = float(split_line[area_index + 3])
        component.area_ref_comp_id = int(split_line[area_index + 4])
        component.area_ref_comp_factor = float(split_line[area_index + 4])

        width_index = split_line.index("MFWHM")
        component.width = float(split_line[width_index + 1])
        component.width_min = float(split_line[width_index + 2])
        component.width_max = float(split_line[width_index + 3])
        component.width_ref_comp_id = int(split_line[width_index + 4])
        component.width_ref_comp_factor = float(split_line[width_index + 5])

        position_index = split_line.index("Position")
        component.position = float(split_line[position_index + 1])
        component.position_min = float(split_line[position_index + 2])
        component.position_max = float(split_line[position_index + 3])
        component.position_ref_comp_id = int(split_line[position_index + 4])
        component.position_ref_comp_factor = float(split_line[position_index + 5])

        rsf_index = split_line.index("RSF")
        component.rsf = float(split_line[rsf_index + 1])
        mass_index = split_line.index("MASS")
        component.mass = float(split_line[mass_index + 1])
        comp_index = split_line.index("INDEX")
        component.index = int(split_line[comp_index + 1])
        component.tag = str(split_line[comp_index + 2])
        const_index = split_line.index("CONST")
        component.const = str(split_line[const_index + 1])
        uncorrected_rsf_index = split_line.index("UNCORRECTEDRSF")
        component.uncorrected_rsf = float(split_line[uncorrected_rsf_index + 1])

        component.validate_types()

        self.casa_data["components"] += [component]

    def _split_line(self, line):
        """Split line in Casa processing."""
        regex_pattern = r"\s+|\(\*(?:(?!\*\)).)*?\*\)|\S+"
        result = [match for match in re.findall(regex_pattern, line) if match.strip()]
        return [re.sub(r"[\(*?\*)]", "", text) for text in result]


@dataclass
class CasaEnergyCalibration(XpsDataclass):
    """An object to store one CasaXPS energy calibration."""

    energy_type: str = "binding"
    energy_offset: float = 0.0
    energy_offset_units: str = "eV"
    measured_energies: list = field(default_factory=list)
    measured_energies_units: str = "eV"
    aligned_energy: float = 0.0
    aligned_energy_units: str = "eV"
    operation: str = "ADD"
    range_calibration: bool = False

    def apply_energy_shift(self, x: float):
        if self.operation == "addition":
            if self.energy_type == "binding":
                return x - self.energy_offset

            elif self.energy_type == "kinetic":
                return x + self.energy_offset

        if self.range_calibration:
            pass  # ToDo: apply range calibration


@dataclass
class CasaIntensityCalibration(XpsDataclass):
    """An object to store one CasaXPS energy calibration."""

    # TODO: enable more types of intensity calibrations
    power_law_coefficient: float = 0.0
    tf_function_ordinate_no: int = 0


@dataclass
class CasaRegion(XpsDataclass):
    """An object to store one CasaXPS peak fit region."""

    name: str = ""
    rsf: float = 0.0
    start: float = 0.0
    start_units: str = "eV"
    end: float = 0.0
    end_units: str = "eV"
    bg_type: str = ""
    av_width: float = 0.0
    av_width_units: str = "eV"
    start_offset: float = 0.0
    start_offset_units: str = ""
    end_offset: float = 0.0
    end_offset_units: str = ""
    cross_section: list = field(default_factory=list)
    tag: str = ""
    unknown_0: float = 0.0
    unknown_1: float = 0.0
    rsf_effective: float = 0.0

    def calculate_background(self, x: np.ndarray, y: np.ndarray):
        backgrounds: Dict[str, Any] = {
            "Linear": LinearBackground,
            "Shirley": Shirley,
            "Step Up": StepUp,
            "Step Down": StepDown,
            "U 2 Tougaard": TougaardU3,
            "U 3 Tougaard": TougaardU3,
            "U 4 Tougaard": TougaardU4,
        }

        background_params = [
            float(param) for param in self.cross_section if float(param)
        ]

        try:
            background_class = backgrounds[self.bg_type]

            try:
                background = background_class(*background_params)
            except TypeError:
                background = background_class()

            if self.end > self.start:
                min_x = self.start
                max_x = self.end
            else:
                min_x = self.end
                max_x = self.start

            region = np.argwhere((x >= min_x) & (x <= max_x))
            fit_region = slice(region[0, 0], region[-1, 0], 1)

            self.start_offset = 100

            y_start_offset = y[0] * (self.start_offset / 100.0)
            y_end_offset = y[-1] * (self.end_offset / 100.0)
            y[0] -= y_start_offset
            y[-1] -= y_end_offset

            x, y = x[fit_region], y[fit_region]

            self.data = background.calc_background(x, y)

            self.formula = background.formula()
            self.description = str(background)

        except KeyError:
            logger.warning(
                f"Background {self.name} (type {self.bg_type}) could not be parsed because no model exists for it."
            )


@dataclass
class CasaComponent(XpsDataclass):
    """An object to store one CasaXPS peak fit component."""

    name: str = ""
    index: int = -1
    lineshape: str = ""

    area: float = 0.0
    area_min: float = 0.0
    area_max: float = 0.0
    area_ref_comp_id: int = -1
    area_ref_comp_factor: float = 0.0

    width: float = 0.0
    width_units: str = "eV"
    width_min: float = 0.0
    width_min_units: str = "eV"
    width_max: float = 0.0
    width_max_units: str = "eV"
    width_ref_comp_id: int = -1
    width_ref_comp_factor: float = 0.0

    position: float = 0.0
    position_units: str = "eV"
    position_min: float = 0.0
    position_min_units: str = "eV"
    position_max: float = 0.0
    position_max_units: str = "eV"
    position_ref_comp_id: int = -1
    position_ref_comp_factor: float = 0.0

    rsf: float = 0.0
    uncorrected_rsf: float = 0.0
    mass: float = 0.0
    tag: str = ""
    const: str = ""  # CONST

    atomic_concentration: float = 0.0

    def calculate_lineshape(self, x: np.ndarray):
        lineshapes: Dict[str, Any] = {
            "GL": GaussianLorentzianProduct,
            "SGL": GaussianLorentzianSum,
            "LA": LorentzianAsymmetric,
            "LF": LorentzianFinite,
            "DS": DoniachSunjic,
        }

        leading_letters, params = split_after_letters(self.lineshape)

        try:
            peak_class = lineshapes[leading_letters]

            peak_parameters = [self.position, self.width, self.area] + params

            peak = peak_class(*peak_parameters)

            self.data = peak.calc_lineshape(x)
            self.formula = peak.formula()
            self.description = str(peak)

        except KeyError:
            logger.warning(
                f"Component {self.name} (index {self.index}) could not be parsed because no model exists for it."
            )



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Parser for reading XPS (X-ray Photoelectron Spectroscopy) data from
Specs Lab Prodigy XML export, to be passed to mpes nxdl
(NeXus Definition Language) template.
"""

import re
import xml.etree.ElementTree as EmtT
from typing import Tuple, List, Dict, Any
import copy
import xarray as xr
import numpy as np

from pynxtools_xps.reader_utils import XPSMapper, align_name_part
from pynxtools_xps.value_mappers import convert_energy_scan_mode

KEY_PARTS = ["RegionGroup_", "RegionData_"]

KEY_PATTERNS = [re.compile(rf"{key_part}(.*?)(?=\/|$)") for key_part in KEY_PARTS]


def construct_entry_name(key: str) -> str:
    """Construct entry name."""
    name_parts = []

    for key_pattern in KEY_PATTERNS:
        match = re.search(key_pattern, key)
        if match:
            name_part = align_name_part(match.group(1))
            name_parts.append(name_part)
    return "__".join(name_parts)


VALUE_MAP = {
    "scan_mode/name": convert_energy_scan_mode,
}


class XmlMapperSpecs(XPSMapper):
    """
    Class for restructuring xml data file from
    specs vendor into python dictionary.
    """

    config_file = "config_specs_xml.json"

    def _select_parser(self):
        """
        Select Specs XML parser.
        Currently, there is only one parser.
        Returns
        -------
        XmlParserSpecs
            Parser for reading .xml file exported by SpecsLab2.

        """
        return XmlParserSpecs()

    def parse_file(self, file, **kwargs):
        """
        Parse the file using the parser that fits the Prodigy SLE version.
        Returns flat list of dictionaries containing one spectrum each.

        """
        self.file = file
        self.parser = self._select_parser()
        metadata_dict, self.raw_data = self.parser.parse_file(file, **kwargs)

        self._xps_dict = {**self._xps_dict, **metadata_dict}

        self.construct_data()

        return self.data_dict

    # pylint: disable=too-many-locals,too-many-statements, too-many-branches
    def construct_data(self):
        """
        Construct the Binding Energy and separate the counts for
        different detectors and finally sum up all the counts for
        to find total electron counts.
        """
        spectra = copy.deepcopy(self.raw_data)
        self._xps_dict["data"]: dict = {}

        for entry, spectrum in spectra.items():
            raw_data = spectrum["raw_data"]
            mcd_num = int(raw_data["mcd_num"])

            curves_per_scan = raw_data["curves_per_scan"]
            values_per_curve = raw_data["values_per_curve"]
            values_per_scan = int(curves_per_scan * values_per_curve)
            mcd_head = int(raw_data["mcd_head"])
            mcd_tail = int(raw_data["mcd_tail"])
            excitation_energy = raw_data["excitation_energy"]
            scan_mode = raw_data["scan_mode"]
            kinetic_energy = raw_data["kinetic_energy"]
            scan_delta = raw_data["scan_delta"]
            pass_energy = raw_data["pass_energy"]
            kinetic_energy_base = raw_data["kinetic_energy_base"]
            # Adding one unit to the binding_energy_upper is added as
            # electron comes out if energy is one unit higher
            binding_energy_upper = (
                excitation_energy - kinetic_energy + kinetic_energy_base + 1
            )

            mcd_energy_shifts = raw_data["mcd_shifts"]
            mcd_energy_offsets = []
            offset_ids = []

            # consider offset values for detector with respect to
            # position at +16 which is usually large and positive value
            for mcd_shift in mcd_energy_shifts:
                mcd_energy_offset = (mcd_energy_shifts[-1] - mcd_shift) * pass_energy
                mcd_energy_offsets.append(mcd_energy_offset)
                offset_id = round(mcd_energy_offset / scan_delta)
                offset_ids.append(int(offset_id - 1 if offset_id > 0 else offset_id))

            # Skiping entry without count data
            if not mcd_energy_offsets:
                continue
            mcd_energy_offsets = np.array(mcd_energy_offsets)
            # Putting energy of the last detector as a highest energy
            starting_eng_pnts = binding_energy_upper - mcd_energy_offsets
            ending_eng_pnts = starting_eng_pnts - values_per_scan * scan_delta

            channeltron_eng_axes = np.zeros((mcd_num, values_per_scan))
            for ind in np.arange(len(channeltron_eng_axes)):
                channeltron_eng_axes[ind, :] = np.linspace(
                    starting_eng_pnts[ind], ending_eng_pnts[ind], values_per_scan
                )

            channeltron_eng_axes = np.round(channeltron_eng_axes, decimals=8)
            # construct ultimate or incorporated energy axis from
            # lower to higher energy
            scans = list(raw_data["scans"].keys())

            # Check whether array is empty or not
            if not scans:
                continue
            if not raw_data["scans"][scans[0]].any():
                continue
            # Sorting in descending order
            binding_energy = channeltron_eng_axes[-1, :]

            self._xps_dict["data"][entry] = xr.Dataset()

            for scan_nm in scans:
                channel_counts = np.zeros((mcd_num + 1, values_per_scan))
                # values for scan_nm corresponds to the data for each
                # "scan" in individual CountsSeq
                scan_counts = raw_data["scans"][scan_nm]

                if scan_mode == "fixed_analyser_transmission":
                    for row in np.arange(mcd_num):
                        count_on_row = scan_counts[row::mcd_num]
                        # Reverse counts from lower to higher
                        # BE as in BE_eng_axis
                        count_on_row = count_on_row[mcd_head:-mcd_tail]

                        channel_counts[row + 1, :] = count_on_row
                        channel_counts[0, :] += count_on_row

                        # Storing detector's raw counts
                        self._xps_dict["data"][entry][f"{scan_nm}_chan{row}"] = (
                            xr.DataArray(
                                data=channel_counts[row + 1, :],
                                coords={"energy": binding_energy},
                            )
                        )

                        # Storing callibrated and after accumulated each scan counts
                        if row == mcd_num - 1:
                            self._xps_dict["data"][entry][scan_nm] = xr.DataArray(
                                data=channel_counts[0, :],
                                coords={"energy": binding_energy},
                            )
                else:
                    for row in np.arange(mcd_num):
                        start_id = offset_ids[row]
                        count_on_row = scan_counts[start_id::mcd_num]
                        count_on_row = count_on_row[0:values_per_scan]
                        channel_counts[row + 1, :] = count_on_row

                        # shifting and adding all the curves.
                        channel_counts[0, :] += count_on_row

                        # Storing detector's raw counts
                        self._xps_dict["data"][entry][f"{scan_nm}_chan{row}"] = (
                            xr.DataArray(
                                data=channel_counts[row + 1, :],
                                coords={"energy": binding_energy},
                            )
                        )

                        # Storing callibrated and after accumulated each scan counts
                        if row == mcd_num - 1:
                            self._xps_dict["data"][entry][scan_nm] = xr.DataArray(
                                data=channel_counts[0, :],
                                coords={"energy": binding_energy},
                            )

            # Write averaged cycle data to 'data'.
            averaged_scans = {
                key: value
                for key, value in self._xps_dict["data"][entry].items()
                if "_chan" not in key
            }

            average_cycles = {}

            for key, value in averaged_scans.items():
                cycle_number = re.findall(r"cycle(\d+)_", key)[0]
                cycle_key = f"cycle{cycle_number}"
                if cycle_number not in average_cycles:
                    average_cycles[cycle_key] = value
                else:
                    average_cycles[cycle_key] += value

            for cycle, value in average_cycles.items():
                average_cycles[cycle] = value / len(
                    [k for k in averaged_scans.keys() if cycle in k]
                )

            for cycle, value in average_cycles.items():
                self._xps_dict["data"][entry][cycle] = xr.DataArray(
                    data=value,
                    coords={"energy": binding_energy},
                )


class XmlParserSpecs:
    """Parser for SpecsLab2 XML data"""

    def __init__(self) -> None:
        self.metadata_dict: dict = {}
        self.entry_to_data: dict = {}
        self._root_path = "/ENTRY"
        self.tail_part_frm_struct = ""
        self.tail_part_frm_othr = ""
        self.child_nm_reslvers = "__child_name_resolver__"

    def parse_file(self, file: str, **kwargs):
        """Start parsing process and parse children recursively.

        Parameters
        ----------
        """
        root_element = EmtT.parse(file).getroot()
        root_element.attrib[self.child_nm_reslvers] = []  # type: ignore[assignment]
        child_num = len(root_element)
        parent_path = self._root_path
        skip_child = -1

        child_elmt_ind = 0
        while child_num > 0:
            self.pass_child_through_parsers(
                root_element, parent_path, child_elmt_ind, skip_child
            )

            child_num -= 1
            child_elmt_ind += 1

        metadata_dict: Dict[str, Any] = {}

        for key, value in self.metadata_dict.items():
            entry_name = construct_entry_name(key)
            str_to_replace = "/"
            for i, key_pattern in enumerate(KEY_PATTERNS):
                match = re.search(key_pattern, key)
                if match:
                    if KEY_PARTS[i] == "RegionData_":
                        str_to_replace += "regions/"
                    str_to_replace += f"{KEY_PARTS[i]}{match.group(1)}/"
            modified_key = key.replace(str_to_replace, f"[{entry_name}]/")

            for suffix, func in VALUE_MAP.items():
                if modified_key.endswith(suffix):
                    value = func(value)

            metadata_dict[modified_key] = value

        self.metadata_dict = metadata_dict

        self.collect_raw_data_to_construct_data()

        return self.metadata_dict, self.entry_to_data

    def pass_child_through_parsers(
        self,
        element_: EmtT.Element,
        parent_path: str,
        child_elmt_ind: int,
        skip_child: int,
    ) -> None:
        """
        Parse the element to parser according to element tag.
        Parameters
        ----------
        element_ : xml element to parse
        parent_path : Xpath of the parent element where the element_ belongs
        child_elmt_ind : Index of child element to track the children.
        skip_child : Tack the children who will be skipped to pass to
                     the parser
        Returns
        -------
        None
        """

        name_val_elmt_tag = ["ulong", "double", "string", "boolean", "enum", "any"]

        parent_element = element_
        element = parent_element[child_elmt_ind]
        element.attrib["__parent__"] = parent_element  # type: ignore[assignment]
        element.attrib["__odr_siblings__"] = child_elmt_ind  # type: ignore[assignment]

        if self.child_nm_reslvers not in parent_element.attrib.keys():
            parent_element.attrib[self.child_nm_reslvers] = []  # type: ignore[assignment]

        elmt_tag = element.tag

        if child_elmt_ind <= skip_child:
            pass

        elif elmt_tag == "sequence":
            self.parse_sequence(element, parent_path)

        elif elmt_tag == "struct":
            self.parse_struct(element, parent_path)

        elif elmt_tag in name_val_elmt_tag:
            self.last_element_parser(element, parent_path)

        else:
            raise TypeError("Needs to parse to different type of parser")

    def parse_sequence(self, element_: EmtT.Element, parent_path: str) -> None:
        """
        Parameters
        ----------
        element_ : Element with sequence tag
        parent_path : Xpath of the parent element where the element_ belongs

        Returns
        -------
        None
        """

        child_num = len(element_)
        elmt_attr = element_.attrib

        section_nm_reslvr = ""
        key_name = "name"
        if key_name in elmt_attr.keys():
            section_nm_reslvr = f"{elmt_attr[key_name]}"
            section_nm_reslvr = self.check_for_siblings_with_same_name(
                section_nm_reslvr, element_
            )

            parent_path = f"{parent_path}/{section_nm_reslvr}"

        child_elmt_ind = 0
        while child_num > 0:
            self.pass_child_through_parsers(
                element_, parent_path, child_elmt_ind, skip_child=-1
            )
            child_num -= 1
            child_elmt_ind += 1

    def struct_fc_name_sc_value(self, element_, first_child, parent_path, skip_child):
        """Struct representing parameter with first child (fc) 'name'
        and second child(sc) 'value'."""
        section_nm_reslvr = ""
        units = ["mV", "degree", "W", "kV", "ns"]

        section_nm_reslvr = self.restructure_value(first_child.text, first_child.tag)
        section_nm_reslvr = self.check_for_siblings_with_same_name(
            section_nm_reslvr, element_
        )
        skip_child += 1
        # Separating the units
        for unit in units:
            if f"_[{unit}]" in section_nm_reslvr:
                section_nm_reslvr, _ = section_nm_reslvr.split("_")
                self.metadata_dict[f"{parent_path}/{section_nm_reslvr}/@units"] = unit

        parent_path, self.tail_part_frm_struct = self.check_last_part_repetition(
            parent_path, self.tail_part_frm_struct, section_nm_reslvr
        )

        return parent_path, skip_child

    def struct_fc_name_sc_string(self, element_, first_child, parent_path, skip_child):
        """Struct representing parameter with first child(fc) 'name'
        and first child(fc) having 'string'."""

        elmt_attr = element_.attrib
        key_type_name = "type_name"
        skip_child += 1
        child_txt = self.restructure_value(first_child.text, first_child.tag)

        section_nm_reslvr = f"{elmt_attr[key_type_name]}_{child_txt}"
        section_nm_reslvr = self.check_for_siblings_with_same_name(
            section_nm_reslvr, element_
        )

        parent_path = f"{parent_path}/{section_nm_reslvr}"

        return parent_path, skip_child

    def parse_struct(self, element_: EmtT.Element, parent_path: str) -> None:
        """
        Parameters
        ----------
        element_ : Element with struct tag
        parent_path : Xpath of the parent element where the element_ belongs

        Returns
        -------
        None
        """

        child_num = len(element_)
        elmt_attr = element_.attrib

        # Resolving struct name section is here
        skip_child = -1
        section_nm_reslvr = ""
        first_child = element_[0]
        second_child = element_[1]

        key_name = "name"
        key_value = "value"
        key_type_name = "type_name"
        if key_name in elmt_attr.keys():
            section_nm_reslvr = elmt_attr[key_name]
            section_nm_reslvr = self.check_for_siblings_with_same_name(
                section_nm_reslvr, element_
            )
            parent_path, self.tail_part_frm_struct = self.check_last_part_repetition(
                parent_path, self.tail_part_frm_struct, section_nm_reslvr
            )

        elif key_name not in elmt_attr.keys():
            if (
                key_name in first_child.attrib.values()
                and key_value in second_child.attrib.values()
            ):
                parent_path, skip_child = self.struct_fc_name_sc_value(
                    element_, first_child, parent_path, skip_child
                )
            elif (
                key_name in first_child.attrib.values() and first_child.tag == "string"
            ):
                parent_path, skip_child = self.struct_fc_name_sc_string(
                    element_, first_child, parent_path, skip_child
                )

            else:
                # Check twin siblings
                section_nm_reslvr = self.restructure_value(
                    elmt_attr[key_type_name], "string"
                )
                section_nm_reslvr = (
                    section_nm_reslvr + "_" + str(elmt_attr["__odr_siblings__"])
                )
                parent_path = f"{parent_path}/{section_nm_reslvr}"

        child_elmt_ind = 0
        while child_num > 0:
            self.pass_child_through_parsers(
                element_, parent_path, child_elmt_ind, skip_child
            )
            child_num -= 1
            child_elmt_ind += 1

    def last_element_parser(self, element_: EmtT.Element, parent_path: str) -> None:
        """

        Parameters
        ----------
        element_ : Element with a tag among 'ulong', 'double', 'string',
                   'boolean', 'enum', 'any'
        parent_path : Xpath of the parent element where the element_ belongs

        Returns
        -------
        None
        """

        child_num = len(element_)
        elmt_attr = element_.attrib

        if child_num == 0:
            if "name" in elmt_attr.keys():
                section_nm_reslvr = f"{elmt_attr['name']}"
                value = self.restructure_value(element_.text, element_.tag)

                parent_path, self.tail_part_frm_othr = self.check_last_part_repetition(
                    parent_path, self.tail_part_frm_othr, section_nm_reslvr
                )
                self.metadata_dict[f"{parent_path}"] = value
            else:
                self.metadata_dict[f"{parent_path}"] = self.restructure_value(
                    element_.text, element_.tag
                )
        elif child_num == 1 and "any" == element_.tag:
            child_elmt = element_[0]
            self.metadata_dict[f"{parent_path}"] = self.restructure_value(
                child_elmt.text, child_elmt.tag
            )

    def check_for_siblings_with_same_name(self, reslv_name, new_sblings_elmt):
        """Check for the same name in the same level. For elments with the same
        write the name _1, _2... .
        """
        child_nm_reslvr_li = new_sblings_elmt.attrib["__parent__"].attrib[
            self.child_nm_reslvers
        ]
        if reslv_name not in child_nm_reslvr_li:
            parent = new_sblings_elmt.attrib["__parent__"]
            parent.attrib[self.child_nm_reslvers].append(reslv_name)
        else:
            last_twin_sib_nm = child_nm_reslvr_li[-1]
            try:
                ind = last_twin_sib_nm.split("_")[-1]
                reslv_name = f"{reslv_name}_{int(ind) + 1}"
                parent = new_sblings_elmt.attrib["__parent__"]
                parent.attrib[self.child_nm_reslvers].append(reslv_name)
            except ValueError:
                reslv_name = f"{reslv_name}_1"
                parent = new_sblings_elmt.attrib["__parent__"]
                parent.attrib[self.child_nm_reslvers].append(reslv_name)
        return reslv_name

    def check_last_part_repetition(
        self, parent_path: str, pre_tail_part: str, new_tail_part: str
    ) -> Tuple[str, str]:
        """
         Check for the data from the same group, for example repetition of the
         experiments under the same physical circumstances, make number of
         them with ..._1, ..._2.
        Parameters
        ----------
        self : XmlSpecs object
        parent_path : Xpath of the parent element
        pre_tail_part : The tail part added in the previous step
        new_tail_part : The tail part obtained from present element

        Returns
        -------
        parent_path : Newly obtained or replaced parent Xpath
        pre_tail_part : Newly obtained or replaced tail_path
        """
        if new_tail_part == pre_tail_part:
            previous_key = f"{parent_path}/{new_tail_part}"
            previous_val = self.metadata_dict.get(previous_key, None)
            if previous_val:
                self.metadata_dict[f"{parent_path}/{new_tail_part}_0"] = previous_val
                pre_tail_part = f"{new_tail_part}_1"
                parent_path = f"{parent_path}/{pre_tail_part}"

                del self.metadata_dict[previous_key]

                return parent_path, pre_tail_part

            parent_path = f"{parent_path}/{pre_tail_part}"
            return parent_path, pre_tail_part

        if new_tail_part in pre_tail_part:
            try:
                ind_ = pre_tail_part.split("_")[-1]
                ind = int(ind_)
                pre_tail_part = f"{new_tail_part}_{ind + 1}"
                parent_path = f"{parent_path}/{pre_tail_part}"
                return parent_path, pre_tail_part
            except ValueError:
                parent_path = f"{parent_path}/{new_tail_part}"
                pre_tail_part = new_tail_part

                return parent_path, pre_tail_part
            except TypeError:
                parent_path = f"{parent_path}/{new_tail_part}"
                pre_tail_part = new_tail_part

                return parent_path, pre_tail_part

        parent_path = f"{parent_path}/{new_tail_part}"
        pre_tail_part = new_tail_part

        return parent_path, pre_tail_part

    @staticmethod
    def restructure_value(value_text: str, element_tag: str) -> Any:
        """
            Collect the value_text transform it to the data_type according
            to the type name provided by element_tag.
        Parameters
        ----------
        value_text : text data that would be 'unsigned long', 'double', 'string',
                    'boolean', 'enum/string'
        element_tag : tag name among 'unsigned long', 'double', 'string',
                    'boolean', 'enum/string'

        Returns
        -------

        """

        def double_(para):
            return np.double(para)

        def ulong_(para):
            return np.uint(para)

        def bool_(para):
            return np.bool_(para)

        data_ty = {"double": double_, "ulong": ulong_, "boolean": bool_}
        string_ty = ["string", "enum"]

        if not value_text:
            return ""

        if element_tag in string_ty:
            value_text_: Any = " ".join(value_text.split()).replace(" ", "_")
            return value_text_

        for key_, _ in data_ty.items():
            if key_ == element_tag:
                value_text_ = value_text.split()
                numpy_value = data_ty[element_tag](value_text_)[...]
                if np.shape(numpy_value) == (1,):
                    return numpy_value[0]
                return numpy_value

    def cumulate_counts_series(
        self,
        scan_seq_elem: EmtT.Element,
        counts_length: int = None,
        cumulative_counts: np.ndarray = None,
    ) -> Tuple[str, np.ndarray]:
        """
        Sum the counts over different scans. Each ScanSeaq contains
        multiple scans under the same physical environment. The
        multiple scans are usually taken to make the peaks visible and
        distinguishable.

        Parameters
        ----------
        scan_seq_elem : Element with ScanSeq tag
        counts_length : Number of count (length of 1D numpy array)
                        contain in each scan
        cumulative_counts : Cumulative counts up to last scan  from the
                            same ScanSeq

        Returns
        -------
        np.ndarray : Cumulative up to last scans from the same ScanSeq
        """

        child_num = len(scan_seq_elem)
        name = "count"

        child_elmt_ind = 0
        while child_num >= 0:
            if scan_seq_elem.attrib["type_name"] == "CountsSeq":
                num_of_counts = int(scan_seq_elem.attrib["length"])
                if not counts_length:
                    counts_length = num_of_counts
                if counts_length != num_of_counts:
                    raise ValueError("Count number from all scans must be equal!!")

            if scan_seq_elem.attrib["type_name"] == "Counts":
                counts_data = self.restructure_value(
                    scan_seq_elem.text, scan_seq_elem.tag
                )

                if cumulative_counts is None:
                    cumulative_counts = counts_data
                else:
                    cumulative_counts = cumulative_counts + counts_data

            if child_num > 0:
                child_element = scan_seq_elem[child_elmt_ind]
                name, cumulative_counts = self.cumulate_counts_series(
                    child_element, counts_length, cumulative_counts
                )

            child_num = child_num - 1
            child_elmt_ind = child_elmt_ind + 1

        return (name, cumulative_counts)

    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    def collect_raw_data_to_construct_data(self):
        """Collect the raw data about detectors so that the binding energy and
        and counts for the corresponding nominal.
        """

        entry_list: List = []
        raw_dict = {
            "mcd_num": 0,
            "curves_per_scan": 0,
            "values_per_curve": 0,
            "mcd_head": 0,
            "mcd_tail": 0,
            "excitation_energy": 0,
            "kinetic_energy": 0,
            "kinetic_energy_base": 0,
            "effective_workfunction": 0,
            "scan_delta": 0,
            "pass_energy": 0,
            "mcd_shifts": [],
            "mcd_poss": [],
            "mcd_gains": [],
            "time": 0,
            "scans": {},
        }
        for key, val in self.metadata_dict.items():
            if not key.endswith("parameters/File"):
                entry = construct_entry_name(key)
                match = re.search(r"/ENTRY\[([^\]]+)\]", key)
                if match:
                    entry = match.group(1)
                else:
                    entry = ""

            if entry and (entry not in entry_list):
                self.entry_to_data[entry] = {"raw_data": copy.deepcopy(raw_dict)}
                entry_list.append(entry)

            if "region/curves_per_scan" in key:
                self.entry_to_data[entry]["raw_data"]["curves_per_scan"] = val
            elif "region/values_per_curve" in key:
                self.entry_to_data[entry]["raw_data"]["values_per_curve"] = val

            elif "region/excitation_energy" in key:
                self.entry_to_data[entry]["raw_data"]["excitation_energy"] = val

            elif "region/scan_mode/name" in key:
                val = convert_energy_scan_mode(val)
                self.entry_to_data[entry]["raw_data"]["scan_mode"] = val

            elif "region/kinetic_energy" in key:
                if "region/kinetic_energy_base" not in key:
                    self.entry_to_data[entry]["raw_data"]["kinetic_energy"] = val
                    continue
                if "region/kinetic_energy_base" in key:
                    self.entry_to_data[entry]["raw_data"]["kinetic_energy_base"] = val
                    continue

            elif "region/effective_workfunction" in key:
                self.entry_to_data[entry]["raw_data"]["effective_workfunction"] = val

            elif "region/scan_delta" in key:
                self.entry_to_data[entry]["raw_data"]["scan_delta"] = val

            elif "region/pass_energy" in key:
                self.entry_to_data[entry]["raw_data"]["pass_energy"] = val

            elif "mcd_head" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_head"] = val

            elif "mcd_tail" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_tail"] = val

            elif "shift" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_shifts"].append(val)
                self.entry_to_data[entry]["raw_data"]["mcd_num"] += 1

            elif "gain" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_gains"].append(val)

            elif "position" in key:
                self.entry_to_data[entry]["raw_data"]["mcd_poss"].append(val)

            # construct scan names e.g cycles2_scan0
            if "cycles/Cycle_" in key:
                _, last_part = key.split("cycles/Cycle_")
                if "/time" in last_part:
                    self.entry_to_data[entry]["raw_data"]["time"] = val
                    continue
                if "/parameters/Loop" in last_part:
                    self.entry_to_data[entry]["raw_data"]["loop_no"] = val
                    continue
                parts = last_part.split("/")
                cycle_num, scan_num = parts[0], parts[-2].split("_")[1]
                scan_name = f"cycle{cycle_num}_scan{scan_num}"

                self.entry_to_data[entry]["raw_data"]["scans"][scan_name] = val



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-instance-attributes
"""
Data model for Vamas ISO standard.
"""

from dataclasses import dataclass, field
from pynxtools_xps.reader_utils import XpsDataclass


@dataclass(order=True)
class VamasHeader(XpsDataclass):
    """An object to store the Vamas header information."""

    format_id: str = (
        "VAMAS Surface Chemical Analysis Standard Data Transfer Format 1988 May 4"
    )
    institute_id: str = "Not Specified"
    instrument_model_id: str = "Not Specified"
    operator_id: str = "Not Specified"
    experiment_id: str = "Not Specified"
    num_comment_lines: int = 0
    comment_lines: list = field(
        default_factory=list
    )  # ["Casa Info Follows CasaXPS Version 2.3.22PR1.0\n0"]
    exp_mode: str = "NORM"
    scan_mode: str = "REGULAR"
    num_spectral_regions: int = 0
    num_analysis_positions: int = 0
    num_x_coords: int = 0
    num_y_coords: int = 0
    num_exp_var: int = 1
    num_entries_in_inclusion_list: int = 0
    inclusion_list: list = field(default_factory=list)
    num_manually_entered_items_in_block: int = 0
    manually_entered_items_in_block: list = field(default_factory=list)
    num_future_upgrade_exp_entries: int = 0
    num_future_upgrade_block_entries: int = 0
    future_upgrade_exp_entries: list = field(default_factory=list)
    num_blocks: int = 1


@dataclass
class VamasBlock(XpsDataclass):
    """An object to store a block of spectrum data and meta-data."""

    block_id: str = ""
    sample_id: str = ""
    year: int = 0
    month: int = 0
    day: int = 0
    hour: int = 0
    minute: int = 0
    second: int = 0
    no_hrs_in_advance_of_gmt: int = 0
    num_comment_lines: int = 0
    # This list should contain one element per for each
    # line in the comment block
    comment_lines: list = field(default_factory=list)
    technique: str = ""
    x_coord: float = 0.0
    y_coord: float = 0.0
    exp_var_value: str = ""
    source_label: str = ""
    sputter_ion_atomic_number: int = 0
    sputter_ion_num_atoms: int = 0
    sputter_ion_charge: int = 1
    source_energy: float = 0.0
    source_power: str = "0"
    source_beam_width_x: str = "0"
    source_beam_width_y: str = "0"
    field_of_view_x: float = 0.0
    field_of_view_y: float = 0.0
    first_linescan_x_start: float = 0.0
    first_linescan_y_start: float = 0.0
    first_linescan_x_end: float = 0.0
    first_linescan_y_end: float = 0.0
    last_linescan_x_end: float = 0.0
    last_linescan_y_end: float = 0.0
    source_polar_angle: float = 0.0
    source_azimuth_angle: float = 180.0
    analyser_mode: str = ""
    resolution: float = 0.0
    differential_width_aes: float = 0.0
    magnification: float = 1.0
    work_function: float = 0.0
    target_bias: float = 0.0
    # analyser slit length divided by the magnification
    # of the analyser transfer lens
    analysis_width_x: float = 0.0
    analysis_width_y: float = 0.0
    # degrees from upward z-direction,
    # defined by the sample stage
    analyser_take_off_polar_angle: float = 0.0
    analyser_take_off_azimuth_angle: float = 0.0
    species_label: str = ""
    transition_label: str = ""
    particle_charge: int = -1
    abscissa_label: str = "kinetic energy"
    abscissa_units: str = "eV"
    abscissa_start: float = 0.0
    abscissa_step: float = 0.0
    no_variables: int = 2
    variable_label_1: str = "counts"
    variable_units_1: str = "d"
    variable_label_2: str = "Transmission"
    variable_units_2: str = "d"
    signal_mode: str = "pulse counting"
    dwell_time: float = 0.0
    no_scans: int = 0
    time_correction: float = 0.0

    sputter_source_energy: float = 0.0
    sputter_source_beam_current: float = 0.0
    sputter_source_width_x: float = 0.0
    sputter_source_width_y: float = 0.0
    sputter_source_incidence_polar_angle: float = 0.0
    sputter_source_azimuth_angle: float = 0.0

    # degrees from upward z-direction,
    # defined by the sample stage
    sample_normal_polar_angle_of_tilt: float = 0.0
    # degrees clockwise from the y-direction towards the
    # operator, defined by the sample stage
    sample_normal_tilt_azimuth_angle: float = 0.0
    sample_rotation_angle: float = 0.0
    no_additional_params: int = 2
    num_ord_values: int = 0
    future_upgrade_block_entries: list = field(default_factory=list)

    def convert_to_binding_energy_scale(self):
        """
        Convert from kinetic to binding energy.

        ToDo: check that components are also converted.
        """
        self.abscissa_label = "binding energy"

        self.x = self.source_energy - self.x
        self.abscissa_start = float(min(self.x))


@dataclass
class ExpVariable(XpsDataclass):
    label: str = ""
    unit: str = ""


@dataclass
class VamasAdditionalParam(XpsDataclass):
    label: str = ""
    unit: str = ""
    value: float = 0.0


@dataclass
class OrdinateValue(XpsDataclass):
    min_ord_value: float = 0.0
    max_ord_value: float = 0.0



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Parser for reading XPS (X-ray Photoelectron Spectroscopy) data from
Phi PHI VersaProbe 4 instruments (.spe or .pro format), to be passed to
MPES nxdl (NeXus Definition Language) template.
"""

import re
import warnings
import copy
import datetime
import logging
import struct
from typing import Any, Dict, List, Union
from pathlib import Path
import pytz
import xarray as xr
import numpy as np

from pynxtools_xps.reader_utils import (
    XPSMapper,
    construct_entry_name,
    safe_arange_with_edges,
    convert_pascal_to_snake,
    split_value_and_unit,
)

from pynxtools_xps.value_mappers import (
    convert_energy_scan_mode,
    convert_measurement_method,
    convert_bool,
    convert_units,
)

from pynxtools_xps.phi.phi_data_model import (
    PhiMetadata,
    PhiSpectralRegion,
    PhiSpatialArea,
)

logger = logging.getLogger(__name__)

SETTINGS_MAP: Dict[str, str] = {
    "FileDesc": "file_description",
    "acq_filename": "acquisition_filename",
    "acq_file_date": "acquisition_file_date",
    "institution": "vendor",
    "operator": "user_name",
    "experiment_i_d": "experiment_id",
    "analyser_work_fcn": "analyser_work_function",
    "analyser_retard_gain": "analyser_retardation_gain",
    "reg_image_interval": "register_image_interval",
    "reg_image_mode": "register_image_mode",
    "reg_image_last": "register_image_last",
    "platen_i_d": "platen_id",
    "s_x_i_filename": "sxi_filename",
    "intensity_recal": "intensity_recalibration",
    "intensity_cal_coeff": "intensity_calibration_coefficients",
    "energy_recal": "energy_recalibration",
    "s_c_a_multiplier_voltage": "sca_multiplier_voltage",
    "C60IonGun": "c60_ion_gun",
    "t_f_c_parameters": "tfc_parameters",
    "image_size_x_y": "image_size",
    "float_volt": "float_voltage",
    "float_enable": "float_enabled",
    "grid_volt": "grid_voltage",
    "condensor_volt": "condenser_lens_voltage",
    "objective_volt": "objective_lens_voltage",
    "bend_volt": "bend_voltage",
    "neutral_float_volt": "neutral_float_voltage",
    "neutral_float_enable": "neutral_float_enabled",
    "neutral_grid_volt": "neutral_grid_voltage",
    "neutral_condensor_volt": "neutral_condenser_lens_voltage",
    "neutral_objective_volt": "neutral_objective_lens_voltage",
    "neutral_bend_volt": "neutral_bend_voltage",
    "no_d_p_data_cyc": "no_depth_profile_cycles",
    "no_pre_sputter_cyc": "no_pre_sputter_cycles",
    "sample_rotation": "profiling_sample_rotation",
    "depth_recal": "profiling_depth_recalibration",
    "sputter_mode": "profiling_sputter_mode",
    "no_depth_reg": "profiling_no_depth_regions",
    "depth_cal_def": "depth_calibration_definition",
    "analyser_mode": "energy_scan_mode",
    "surv_num_cycles": "survey_num_of_cycles",
    "surv_time_per_step": "survey_dwell_time",
    "no_spectral_reg_full": "no_spectral_regions_full",
    "no_spectral_region": "no_spectral_regions",
    "spectral_reg_def_full": "spectral_region_definition_full",
    "spectral_reg_def2_full": "spectral_region_definition2_full",
    "spectral_reg_background_full": "spectral_region_background_full",
    "spectral_reg_hero_full": "spectral_region_hero_full",
    "spectral_reg_i_r_full": "spectral_region_ir_full",
    "no_spectral_reg": "no_spectral_regions",
    "spectral_reg_def": "spectral_region_definition",
    "spectral_reg_def2": "spectral_region_definition2",
    "spectral_reg_background": "spectral_region_background",
    "spectral_reg_hero": "spectral_region_hero",
    "spectral_reg_i_r": "spectral_region_ir",
    "no_spatial_area": "no_spatial_areas",
    "spatial_area_def": "spatial_area_definition",
    "spatial_h_r_photo_cor": "spatial_hr_photo_correction",
    "xray_offset_in_um": "xray_offset",
    "xray_mag_factor": "xray_magnification_factor",
    "xray_rotation_in_deg": "xray_rotation",
    "xray_setting": "xray_settings",
    "neutralizer_current": "flood_gun_current",
    "neutralizer_energy": "flood_gun_energy",
    "flood_gun_filament": "flood_gun_filament_current",
    "flood_gun_extractor": "flood_gun_extractor_voltage",
    "detector _acq _time": "detector_acquisition_time",
    "number _of _channels": "number_of_channels",
    "stage_position": "stage_positions",
    "defect_positioner_i_d": "defect_positioner_id",
    "defect_positioner_aligment": "defect_positioner_alignment",
    "gcib_beam": "gcib_high_voltage",
    "gcib_wien": "gcib_wien_filter_voltage",
    "gcib_bend": "gcib_bend_voltage",
    "gcib_magnet": "gcib_magnet_current",
    "auto_e_gun_neut": "auto_flood_gun",
    "auto_ion_neut": "auto_neutral_ion_source",
}

KEYS_WITH_UNITS: List[str] = [
    "analyser_work_function",
    "source_analyser_angle",
    "analyser_solid_angle",
    "scan_deflection_span_x",
    "scan_deflection_span_y",
    "scan_deflection_offset_x",
    "scan_deflection_offset_y",
    "sca_multiplier_voltage",
    "delay_before_acquire",
    "sputter_current",
    "sputter_rate",
    "sputter_energy",
    "float_voltage",
    "target_sputter_time",
    "sputter_emission",
    "grid_voltage",
    "condenser_lens_voltage",
    "objective_lens_voltage",
    "bend_voltage",
    "deflection_bias",
    "ion_gun_gas_pressure",
    "sputter_emission",
    "deflection_bias",
    "neutral_current",
    "neutral_rate",
    "neutral_energy",
    "neutral_float_voltage",
    "neutral_grid_voltage",
    "neutral_condenser_lens_voltage",
    "neutral_objective_lens_voltage",
    "neutral_bend_voltage",
    "neutral_target_timed_on_time",
    "neutral_emission",
    "neutral_deflection_bias",
    "neutral_ion_gun_gas_pressure",
    "profiling_sputter_delay",
    "survey_dwell_time",
    "xray_anode_power",
    "xray_power",
    "xray_beam_voltage",
    "xray_beam_diameter",
    "xray_condenser_lens_voltage",
    "xray_objective_coil_current",
    "xray_blanking_voltage",
    "xray_filament_current",
    "xray_offset",
    "xray_rotation",
    "xray_emission_current",
    "xray_max_filament_current",
    "xray_stigmator_x",
    "xray_stigmator_y",
    "flood_gun_current",
    "flood_gun_energy",
    "flood_gun_extractor_voltage",
    "flood_gun_filament_current",
    "flood_gun_pulse_length",
    "flood_gun_pulse_frequency",
    "flood_gun_time_per_step",
    "flood_gun_ramp_rate",
    "flood_gun_x_steering",
    "flood_gun_y_steering",
    "sxi_binding_energy",
    "sxi_pass_energy",
    "sxi_lens2_voltage",
    "sxi_lens3_voltage",
    "sxi_lens4_voltage",
    "sxi_lens5_voltage",
    "sxi_rotator",
    "sxi_lens_bias_voltage",
    "sxi_shutter_bias_voltage",
    "detector_acquisition_time",
    "stage_current_rotation_speed",
    "defect_positioner_u",
    "defect_positioner_v",
    "defect_positioner_x",
    "defect_positioner_y",
    "defect_positioner_z",
    "defect_positioner_tilt",
    "defect_positioner_rotation",
    "gcib_sputter_rate",
    "gcib_high_voltage",
    "gcib_ionization",
    "gcib_extractor",
    "gcib_wien_filter_voltage",
    "gcib_bend_voltage",
    "gcib_emission",
    "gcib_magnet_current",
    "gcib_focus",
    "gcib_objective",
    "gcib_focus",
    "gcib_gas_pressure",
    "gcib_cluster_size",
    "gcib_energy_per_atom",
    "deconvolution_pass_energy",
]

UNIT_MISSING: Dict[str, str] = {
    "intensity": "counts_per_second",
    "grid_voltage": "V",
    "condenser_lens_voltage": "V",
    "objective_lens_voltage": "V",
    "bend_voltage": "V",
    "neutral_grid_voltage": "V",
    "neutral_condenser_lens_voltage": "V",
    "neutral_objective_lens_voltage": "V",
    "stage_current_rotation_speed": "degree/s",
    "neutral_bend_voltage": "V",
    "defect_positioner_u": "mm",
    "defect_positioner_v": "mm",
    "defect_positioner_x": "mm",
    "defect_positioner_y": "mm",
    "defect_positioner_z": "mm",
    "defect_positioner_tilt": "degree",
    "defect_positioner_rotation": "degree",
    "flood_gun_pulse_frequency": "1/s",
    "flood_gun_x_steering": "mm",
    "flood_gun_y_steering": "mm",
    "profiling_sputter_delay": "s",
    "scan_deflection_span_x": "mm",
    "scan_deflection_span_y": "mm",
    "scan_deflection_offset_x": "mm",
    "scan_deflection_offset_y": "mm",
    "xray_stigmator_x": "mm",
    "xray_stigmator_y": "mm",
}


class MapperPhi(XPSMapper):
    """
    Class for restructuring .xy data file from
    Phi vendor into python dictionary.
    """

    config_file = "config_phi.json"

    def __init__(self):
        super().__init__()
        self.write_channels_to_data = True

    def _select_parser(self):
        """Select the proper Phi data parser."""
        return PhiParser()

    def construct_data(self):
        """Map Phi format to NXmpes-ready dict."""
        # pylint: disable=duplicate-code
        spectra = copy.deepcopy(self.raw_data)

        self._xps_dict["data"]: dict = {}

        template_key_map: Dict[str, List[str]] = {
            "user": [
                "user_name",
            ],
            "instrument": [
                "bias_box_mode",
                "instrument_model",
                "vendor",
                "source_analyser_angle",
                "source_analyser_angle/@units",
            ],
            "xray_source": [
                "scan_deflection_span_x",
                "scan_deflection_span_x/@units",
                "scan_deflection_span_y",
                "scan_deflection_span_y/@units",
                "scan_deflection_offset_x",
                "scan_deflection_offset_x/@units",
                "scan_deflection_offset_y",
                "scan_deflection_offset_y/@units",
                "xray_anode_material",
                "xray_anode_position",
                "xray_beam_voltage",
                "xray_beam_voltage/@units",
                "xray_blanking_voltage",
                "xray_blanking_voltage/@units",
                "xray_condenser_lens_voltage",
                "xray_condenser_lens_voltage/@units",
                "xray_delay_factor_x",
                "xray_delay_factor_y",
                "xray_emission_control",
                "xray_emission_current",
                "xray_emission_current/@units",
                "xray_filament_current",
                "xray_filament_current/@units",
                "xray_high_power",
                "xray_high_voltage",
                "xray_high_voltage/@units",
                "xray_interlace_interval",
                "xray_magnification_factor_x",
                "xray_magnification_factor_y",
                "xray_max_filament_current",
                "xray_max_filament_current/@units",
                "xray_monochromatized",
                "xray_objective_coil_current",
                "xray_objective_coil_current/@units",
                "xray_offset",
                "xray_offset/@units",
                "xray_power",
                "xray_power/@units",
                "xray_rotation",
                "xray_rotation/@units",
                "xray_step_delay_read_beam",
                "xray_steps_per_diameter",
                "xray_stigmator_x",
                "xray_stigmator_x/@units",
                "xray_stigmator_y",
                "xray_stigmator_y/@units",
            ],
            "beam": [
                "xray_beam_diameter",
                "xray_beam_diameter/@units",
                "xray_energy",
                "xray_energy/@units",
                "xray_spot_size",
                "xray_spot_size/@units",
            ],
            "electronanalyser": [
                "analyser_retardation_gain",
                "analyser_solid_angle",
                "analyser_solid_angle/@units",
                "analyser_work_function",
                "analyser_work_function/@units",
                "sxi_auto_contrast",
                "sxi_auto_contrast_high",
                "sxi_auto_contrast_low",
                "sxi_binding_energy",
                "sxi_binding_energy/@units",
                "sxi_display_mode",
                "sxi_filename",
                "sxi_persistence",
                "sxi_rotator",
                "sxi_rotator/@units",
                "sxi_sec_per_display",
                "sxi_shutter_bias",
                "sxi_shutter_bias_voltage",
                "sxi_shutter_bias_voltage/@units",
            ],
            "collectioncolumn": [
                "narrow_acceptance_angle",
                "sxi_lens2_voltage",
                "sxi_lens2_voltage/@units",
                "sxi_lens3_voltage",
                "sxi_lens3_voltage/@units",
                "sxi_lens4_voltage",
                "sxi_lens4_voltage/@units",
                "sxi_lens5_voltage",
                "sxi_lens5_voltage/@units",
                "sxi_lens_bias_voltage",
                "sxi_lens_bias_voltage/@units",
            ],
            "energydispersion": [
                "energy_scan_mode",
                "pass_energy",
                "pass_energy/@units",
                "xps_scan_mode",
            ],
            "detector": [
                "channel_1_info",
                "channel_2_info",
                "channel_3_info",
                "channel_4_info",
                "channel_5_info",
                "channel_6_info",
                "channel_7_info",
                "channel_8_info",
                "channel_9_info",
                "channel_10_info",
                "channel_11_info",
                "channel_12_info",
                "channel_13_info",
                "channel_14_info",
                "channel_15_info",
                "channel_16_info",
                "channel_17_info",
                "channel_18_info",
                "channel_19_info",
                "channel_20_info",
                "channel_21_info",
                "channel_22_info",
                "channel_23_info",
                "channel_24_info",
                "channel_25_info",
                "channel_26_info",
                "channel_27_info",
                "channel_28_info",
                "channel_29_info",
                "channel_30_info",
                "channel_31_info",
                "channel_32_info",
                "delay_before_acquire",
                "delay_before_acquire/@units",
                "detector_acquisition_time",
                "detector_acquisition_time/@units",
                "number_of_channels",
                "refresh_persistence",
                "dwell_time",
                "dwell_time/@units",
                "sca_multiplier_voltage",
                "sca_multiplier_voltage/@units",
            ],
            "manipulator": [
                "stage_x",
                "stage_x/@units",
                "stage_y",
                "stage_y/@units",
                "stage_z",
                "stage_z/@units",
                "stage_azimuth",
                "stage_azimuth/@units",
                "stage_polar",
                "stage_polar/@units",
                "stage_current_rotation_speed",
                "stage_current_rotation_speed/@units",
            ],
            "defect_positioner": [
                "defect_positioner_alignment",
                "defect_positioner_comment",
                "defect_positioner_id",
                "defect_positioner_reference_image",
                "defect_positioner_rotation",
                "defect_positioner_rotation/@units",
                "defect_positioner_tilt",
                "defect_positioner_tilt/@units",
                "defect_positioner_u",
                "defect_positioner_u/@units",
                "defect_positioner_v",
                "defect_positioner_v/@units",
                "defect_positioner_x",
                "defect_positioner_x/@units",
                "defect_positioner_y",
                "defect_positioner_y/@units",
                "defect_positioner_z",
                "defect_positioner_z/@units",
            ],
            "c60_ion_gun": [
                "c60_ion_gun",
            ],
            "flood_gun": [
                "auto_flood_gun",
                "flood_gun_current",
                "flood_gun_current/@units",
                "flood_gun_energy",
                "flood_gun_energy/@units",
                "flood_gun_extractor_voltage",
                "flood_gun_extractor_voltage/@units",
                "flood_gun_filament_current",
                "flood_gun_filament_current/@units",
                "flood_gun_gain",
                "flood_gun_mode",
                "flood_gun_pulse_frequency",
                "flood_gun_pulse_frequency/@units",
                "flood_gun_pulse_length",
                "flood_gun_pulse_length/@units",
                "flood_gun_ramp_rate",
                "flood_gun_ramp_rate/@units",
                "flood_gun_time_per_step",
                "flood_gun_time_per_step/@units",
                "flood_gun_x_steering",
                "flood_gun_x_steering/@units",
                "flood_gun_y_steering",
                "flood_gun_y_steering/@units",
            ],
            "gcib": [
                "gcib_bend_voltage",
                "gcib_bend_voltage/@units",
                "gcib_cluster_size",
                "gcib_cluster_size/@units",
                "gcib_emission",
                "gcib_emission/@units",
                "gcib_energy_per_atom",
                "gcib_energy_per_atom/@units",
                "gcib_extractor",
                "gcib_extractor/@units",
                "gcib_focus",
                "gcib_focus/@units",
                "gcib_gas_pressure",
                "gcib_gas_pressure/@units",
                "gcib_high_voltage",
                "gcib_high_voltage/@units",
                "gcib_ionization",
                "gcib_ionization/@units",
                "gcib_magnet_current",
                "gcib_magnet_current/@units",
                "gcib_objective",
                "gcib_objective/@units",
                "gcib_raster_offset_x",
                "gcib_raster_offset_x/@units",
                "gcib_raster_offset_y",
                "gcib_raster_offset_y/@units",
                "gcib_raster_size_x",
                "gcib_raster_size_x/@units",
                "gcib_raster_size_y",
                "gcib_raster_size_y/@units",
                "gcib_sputter_rate",
                "gcib_sputter_rate/@units",
                "gcib_wien_filter_voltage",
                "gcib_wien_filter_voltage/@units",
            ],
            "neutral_ion_gun": [
                "auto_neutral_ion_source",
                "neutral_bend_voltage",
                "neutral_bend_voltage/@units",
                "neutral_condenser_lens_voltage",
                "neutral_condenser_lens_voltage/@units",
                "neutral_current",
                "neutral_current/@units",
                "neutral_deflection_bias",
                "neutral_deflection_bias/@units",
                "neutral_emission",
                "neutral_emission/@units",
                "neutral_energy",
                "neutral_energy/@units",
                "neutral_float_enabled",
                "neutral_float_voltage",
                "neutral_float_voltage/@units",
                "neutral_grid_voltage",
                "neutral_grid_voltage/@units",
                "neutral_ion",
                "neutral_ion_gun_gas_pressure",
                "neutral_ion_gun_gas_pressure/@units",
                "neutral_objective_lens_voltage",
                "neutral_objective_lens_voltage/@units",
                "neutral_raster_offset_x",
                "neutral_raster_offset_x/@units",
                "neutral_raster_offset_y",
                "neutral_raster_offset_y/@units",
                "neutral_raster_x",
                "neutral_raster_x/@units",
                "neutral_raster_y",
                "neutral_raster_y/@units",
                "neutral_rate",
                "neutral_rate/@units",
                "neutral_target_timed_on_time",
                "neutral_target_timed_on_time/@units",
            ],
            "sputter_gun": [
                "bend_voltage",
                "bend_voltage/@units",
                "condenser_lens_voltage",
                "condenser_lens_voltage/@units",
                "deflection_bias",
                "deflection_bias/@units",
                "float_enabled",
                "float_voltage",
                "float_voltage/@units",
                "grid_voltage",
                "grid_voltage/@units",
                "ion_gun_gas_pressure",
                "ion_gun_gas_pressure/@units",
                "ion_gun_mode",
                "objective_lens_voltage",
                "objective_lens_voltage/@units",
                "sputter_current",
                "sputter_current/@units",
                "sputter_emission",
                "sputter_emission/@units",
                "sputter_energy",
                "sputter_energy/@units",
                "sputter_ion",
                "sputter_raster_offset_x",
                "sputter_raster_offset_x/@units",
                "sputter_raster_offset_y",
                "sputter_raster_offset_y/@units",
                "sputter_raster_x",
                "sputter_raster_x/@units",
                "sputter_raster_y",
                "sputter_raster_y/@units",
                "sputter_rate",
                "sputter_rate/@units",
                "target_sputter_time",
                "target_sputter_time/@units",
            ],
            "sample": [],
            "process": [
                "deconvolution",
                "deconvolution_pass_energy",
                "deconvolution_pass_energy/@units",
                "energy_recalibration",
                "energy_reference_energy",
                "energy_reference_energy/@units",
                "energy_reference_peak",
                "intensity_calibration_coefficients",
                "intensity_recalibration",
            ],
            "data": [
                "energy",
                "energy/@type",
                "energy/@units",
                "n_values",
                "intensity/@units",
            ],
            "file_info": [
                "experiment_id",
                "file_date",
                "file_description",
                "file_type",
                "photo_filename",
                "software_version",
                "acquisition_file_date",
                "acquisition_filename",
            ],
            "region": [
                "area_definition",
                "area_description",
                "area_hr_photo_correction",
                "area_id",
                "full_region",
                "image_size_x",
                "image_size_y",
                "no_spatial_areas",
                "no_spectral_regions",
                "no_spectral_regions_full",
                "sem_field_of_view",
                "region_background",
                "region_definition",
                "region_definition2",
                "region_hero",
                "region_id",
                "region_ir",
                "survey_num_of_cycles",
                "technique",
                "technique_ex",
                "tfc_parameters",
            ],
            "profiling": [
                "no_depth_profile_cycles",
                "no_pre_sputter_cycles",
                "profiling_sputter_delay",
                "profiling_sputter_delay/@units",
                "profiling_xray_off_during_sputter",
                "profiling_source_blank_during_sputter",
                "profiling_zalar_high_accuracy_interval",
                "profiling_sample_rotation",
                "profiling_depth_recalibration",
                "profiling_sputter_mode",
                "depth_calibration_definition",
                "profiling_no_depth_regions",
            ],
            # "unused": [
            # "peak_to_noise_ratio_state",
            # "platform",
            # "platen_id",
            # "presputter",
            # "register_image",
            # "register_image_interval",
            # "register_image_last",
            # "register_image_mode",
            # ],
        }

        for spectrum in spectra:
            self._update_xps_dict_with_spectrum(spectrum, template_key_map)

    def _update_xps_dict_with_spectrum(
        self, spectrum: Dict[str, Any], key_map: Dict[str, List[str]]
    ):
        """
        Map one spectrum from raw data to NXmpes-ready dict.

        """
        # pylint: disable=too-many-locals,duplicate-code
        entry_parts = []

        for part in ["group_name", "spectrum_type"]:
            val = spectrum.get(part, None)
            if val:
                entry_parts += [val]

        entry = construct_entry_name(entry_parts)
        entry_parent = f"/ENTRY[{entry}]"

        file_parent = f"{entry_parent}/file_info"
        instrument_parent = f"{entry_parent}/instrument"
        analyser_parent = f"{instrument_parent}/electronanalyser"

        path_map: Dict[str, str] = {
            "file_info": f"{file_parent}",
            "user": f"{entry_parent}/user",
            "instrument": f"{instrument_parent}",
            "xray_source": f"{instrument_parent}/xray_source",
            "beam": f"{instrument_parent}/beam",
            "electronanalyser": f"{analyser_parent}",
            "collectioncolumn": f"{analyser_parent}/collectioncolumn",
            "energydispersion": f"{analyser_parent}/energydispersion",
            "detector": f"{analyser_parent}/detector",
            "manipulator": f"{instrument_parent}/manipulator",
            "c60_ion_gun": f"{instrument_parent}/c60_ion_gun",
            "defect_positioner": f"{instrument_parent}/defect_positioner",
            "flood_gun": f"{instrument_parent}/flood_gun",
            "gcib": f"{instrument_parent}/gcib",
            "neutral_ion_gun": f"{instrument_parent}/neutral_ion_gun",
            "sputter_gun": f"{instrument_parent}/sputter_gun",
            "process": f"{entry_parent}/process",
            "sample": f"{entry_parent}/sample",
            "data": f"{entry_parent}/data",
            "region": f"{entry_parent}/region",
            "profiling": f"{entry_parent}/profiling",
        }

        for grouping, spectrum_keys in key_map.items():
            root = path_map[str(grouping)]
            for spectrum_key in spectrum_keys:
                mpes_key = spectrum_key
                self._xps_dict[f"{root}/{mpes_key}"] = spectrum.get(spectrum_key)

        # Create key for writing to data
        cycle_key = "cycle0"

        energy = np.array(spectrum["energy"])

        if entry not in self._xps_dict["data"]:
            self._xps_dict["data"][entry] = xr.Dataset()

        # Write averaged data to 'data'.
        all_scan_data = np.array(
            [np.array(value) for key, value in spectrum["data"].items()]
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            averaged_scans = np.mean(all_scan_data, axis=0)

        self._xps_dict["data"][entry][cycle_key] = xr.DataArray(
            data=averaged_scans,
            coords={"energy": energy},
        )

        # Write scan data to 'data'.
        for scan_no, intensity in spectrum["data"].items():
            xarr_key = f"{cycle_key}_{scan_no}"
            self._xps_dict["data"][entry][xarr_key] = xr.DataArray(
                data=intensity,
                coords={"energy": energy},
            )


class PhiParser:  # pylint: disable=too-few-public-methods
    """
    A parser for reading in PHI VersaProbe 4 data in the .spe or
    .pro format.
    Tested with Software version SS 3.3.3.2.
    """

    def __init__(self):
        """
        Construct the parser.

        """
        self.raw_data: str = ""
        self.spectra: List[Dict[str, Any]] = []

        self.metadata = PhiMetadata()

        self.binary_header_length = 4
        self.spectra_header_length = 24
        self.encoding: List[str, int] = ["<f", 4]

        self.binary_header: np.ndarray = None
        self.spectra_header: np.ndarray = None

        self.value_function_map = {
            "technique": convert_measurement_method,
            "technique_ex": convert_measurement_method,
            "file_type": _map_file_type,
            "file_date": _parse_datetime,
            "acquisition_file_date": _parse_datetime,
            "energy_reference": _convert_energy_referencing,
            "intensity_calibration_coefficients": _map_to_list,
            "energy_recalibration": convert_bool,
            "scan_deflection_span": _map_to_xy,
            "scan_deflection_offset": _map_to_xy,
            "tfc_parameters": _map_to_list,
            "image_size": _map_to_xy,
            "float_enabled": convert_bool,
            "sputter_raster": _map_to_xy_with_units,
            "sputter_raster_offset": _map_to_xy_with_units,
            "neutral_raster": _map_to_xy_with_units,
            "neutral_raster_offset": _map_to_xy_with_units,
            "profiling_xray_off_during_sputter": convert_bool,
            "profiling_source_blank_during_sputter": convert_bool,
            "profiling_depth_recalibration": convert_bool,
            "energy_scan_mode": convert_energy_scan_mode,
            "xray_source": _convert_xray_source_params,
            "xray_stigmator": _map_to_xy,
            "xray_offset": _map_to_xy,
            "xray_magnification_factor": _map_to_xy,
            "xray_delay_factor": _map_to_xy,
            "xray_high_power": convert_bool,
            "xray_emission_control": convert_bool,
            "xray_settings": _convert_xray_source_settings,
            "sxi_auto_contrast": convert_bool,
            "sxi_shutter_bias": convert_bool,
            "stage_positions": _convert_stage_positions,
            "gcib_raster_size": _map_to_xy_with_units,
            "gcib_raster_offset": _map_to_xy_with_units,
            "auto_flood_gun": convert_bool,
            "auto_neutral_ion_source": convert_bool,
            "presputter": convert_bool,
        }

    def parse_file(self, file: Union[str, Path], **kwargs):
        """
        Parse the .spe, .pro file into a list of dictionaries.

        Parsed data is stored in the attribute 'self.data'.
        Each dictionary in the data list is a grouping of related
        attributes. The dictionaries are later re-structured into a
        nested dictionary that more closely resembles the domain logic.

        Parameters
        ----------
        file : str
            XPS data filepath.

        Returns
        -------
        list
            Flat list of dictionaries containing one spectrum each.

        """
        self.raw_data = self._read_lines(file)
        header, data = self._separate_header_and_data()
        # l = b'\x01'.join(data)

        self.parse_header_into_metadata(header)
        regions = self.parse_spectral_regions(header)
        areas = self.parse_spatial_areas(header)

        self.add_regions_and_areas_to_spectra(regions, areas)

        self.parse_binary_header(data)
        self.parse_data_into_spectra(data)

        self.add_metadata_to_each_spectrum()

        return self.spectra

    def _read_lines(self, file: Union[str, Path]):
        """
        Read in all lines from the file as a list of strings.

        Parameters
        ----------
        file : str
            XPS data filepath.

        Returns
        -------
        None.

        """
        with open(file, mode="rb") as txt_file:  # "utf-8"
            data = txt_file.read()

        return data

    def _separate_header_and_data(self):
        """
        Separate header (with metadata) from data.

        Returns
        -------
        None.

        """
        try:  # Windows
            header, data = self.raw_data.split(b"EOFH\r\n")
            header = header.decode("ascii").split("\r")
            header = [line.strip() for line in header if line.strip()]

        except ValueError:  # Linux, MacOS
            header, data = self.raw_data.split(b"EOFH\n")
            header = header.decode("ascii").split("\n")
            header = [line.strip() for line in header if line.strip()]

        return header, data

    def parse_header_into_metadata(self, header: List[str]):
        """
        Parse header into PHiMetadata dataclass.

        Parameters
        ----------
        header : str
            Header data for one spectrum as a String.

        """

        def map_keys(key: str, channel_count: int):
            """Maps key names for better identification of fields."""
            prefix_map = {
                "neut": ("neut_", "neutral_"),
                "prof": ("prof_", "profiling_"),
                "x_ray": ("x_ray", "xray"),
                "egun_neut": ("egun_neut", "flood_gun"),
            }

            for prefix, (initial, replacement) in prefix_map.items():
                if key.startswith(prefix):
                    key = key.replace(initial, replacement)

            if key.startswith("sxi_lens"):
                key += "_voltage"

            if key.startswith("channel _info"):
                key = f"channel_{channel_count}_info"
                channel_count += 1

            key_map: Dict[str, str] = {
                "desc": "description",
                "defect_pos": "defect_positioner",
                "g_c_i_b": "gcib",
            }

            for old_key, new_key in key_map.items():
                if old_key in key:
                    key = key.replace(old_key, new_key)

            key = SETTINGS_MAP.get(key, key)

            return key, channel_count

        channel_count = 1
        datacls_fields = list(self.metadata.__dataclass_fields__.keys())
        datacls_fields = [field for field in datacls_fields if "_units" not in field]

        for line in header:
            try:
                key, value = line.split(": ")

            except ValueError:
                key = line.strip(": ")
                value = ""

            key = convert_pascal_to_snake(key)
            key, channel_count = map_keys(key, channel_count)

            if key in datacls_fields:
                field_type = type(getattr(self.metadata, key))

                if key in KEYS_WITH_UNITS:
                    value, unit = self.extract_unit(key, value)
                    setattr(self.metadata, f"{key}_units", unit)

                if not key.startswith("channel_"):
                    value = self.map_values(key, value, field_type)

                else:
                    value = _convert_channel_info(value)

                setattr(self.metadata, key, value)

        self.metadata.validate_types()

    def parse_spectral_regions(self, header: List[str]):
        """
        Parse spectral regions definitions.

        Parameters
        ----------
        header : list
            List of header strings.

        Returns
        -------
        regions_full : list
            List of regions with `full` keyword.
        regions : list
            List of regions without `full` keyword.

        """
        spectral_defs = [line for line in header if line.startswith("SpectralReg")]

        regions = []
        regions_full = []
        for n in range(self.metadata.no_spectral_regions_full):
            regions_full += [PhiSpectralRegion(region_id=n + 1)]
        for n in range(self.metadata.no_spectral_regions):
            regions += [PhiSpectralRegion(region_id=n + 1)]

        concept_map = {
            "SpectralRegDef": "region_definition",
            "SpectralRegDef2": "region_definition2",
            "SpectralRegBackground": "region_background",
            "SpectralRegHero": "region_hero",
            "SpectralRegIR": "region_ir",
        }

        for region in regions_full:
            region.full_region = True
            for file_key, metadata_key in concept_map.items():
                if spectral_defs and spectral_defs[0].startswith(file_key):
                    setattr(region, metadata_key, spectral_defs[0].split(" ", 1)[1])
                    spectral_defs.pop(0)

        for region in regions:
            region.full_region = False

            for file_key, metadata_key in concept_map.items():
                if spectral_defs and spectral_defs[0].startswith(file_key):
                    setattr(region, metadata_key, spectral_defs[0].split(" ", 1)[1])
                    spectral_defs.pop(0)

        for region in regions:
            def_split = region.region_definition.split(" ")
            region.spectrum_type = def_split[2]
            region.n_values = int(def_split[4])
            step = -float(def_split[5])
            start = float(def_split[6])
            stop = float(def_split[7])
            region.dwell_time = float(def_split[-3])
            region.dwell_time_units = "s"
            region.pass_energy = float(def_split[-2])
            region.energy_type = "binding"
            region.energy_units = "eV"

            region.energy = np.flip(safe_arange_with_edges(stop, start, step))

            region.validate_types()

        return regions

    def parse_spatial_areas(self, header: List[str]):
        """
        Parse spatial areas definitions.

        Parameters
        ----------
        header : list
            List of header strings.

        Returns
        -------
        areas: list
            List of spatial areas and their definitions.

        """
        spatial_defs = [line for line in header if line.startswith("Spatial")]

        areas = []
        for n in range(self.metadata.no_spatial_areas):
            areas += [PhiSpatialArea(area_id=n)]

        concept_map = {
            "SpatialAreaDef": "area_definition",
            "SpatialAreaDesc": "area_description",
            "SpatialHRPhotoCor": "area_hr_photo_correction",
        }

        for area in areas:
            for file_key, metadata_key in concept_map.items():
                if spatial_defs and spatial_defs[0].startswith(file_key):
                    setattr(area, metadata_key, spatial_defs[0].split(" ", 1)[1])
                    spatial_defs.pop(0)

        return areas

    def add_regions_and_areas_to_spectra(self, regions: list, areas: list):
        """
        Define each spectra by its region and area defintions.

        Parameters
        ----------
        regions : list
            List of PhiSpectralRegion objects.
        areas : list
            List of PhiSpatialArea objects

        Returns
        -------
        None.

        """
        for region in regions:
            for area in areas:
                concatenated = {**region.dict(), **area.dict()}

                region_and_areas: Dict[str, Any] = {}

                for key, value in concatenated.items():
                    replacement_map = {
                        "_units": "/@units",
                        "energy_type": "energy/@type",
                    }

                    new_key = key

                    for suffix, replacement in replacement_map.items():
                        if suffix in key:
                            new_key = key.replace(suffix, replacement)

                    region_and_areas[new_key] = value

                self.spectra += [region_and_areas]

    def parse_binary_header(self, binary_data):
        """
        Read the binary headers
        Assuming the headers are 4 bytes unsigned integers
        Each spectrum gets 24 unsigned 4 byte integers

        Parameters
        ----------
        binary_data : bytes
            Binary XPS data, format is 64 bit float.

        """
        binary_header = struct.unpack("I", binary_data[: self.binary_header_length])[0]

        for i, spectrum in enumerate(self.spectra):
            start = (
                self.binary_header_length * self.spectra_header_length * i
                + self.binary_header_length
            )
            stop = start + self.binary_header_length * self.spectra_header_length
            spectrum_header = struct.unpack(
                "I" * self.spectra_header_length, binary_data[start:stop]
            )

            spectrum.update(
                {
                    "binary_header": binary_header,
                    "spectrum_header": np.array(spectrum_header),
                    "n_values": spectrum_header[8],
                    "n_scans": spectrum_header[9],
                    "binary_start": spectrum_header[23],
                    "binary_len": spectrum_header[22],
                }
            )

        self._check_encoding()

    def parse_data_into_spectra(self, binary_data):
        """
        Parse the data of all spectra.

        Parameters
        ----------
        binary_data : bytes
            Binary XPS data, format is 64 bit float.

        """
        for i, spectrum in enumerate(self.spectra):
            float_buffer = self.encoding[1]

            # spectrum["n_scans"] = 2
            spectrum["data"] = {}
            start = spectrum["binary_start"]
            stop = start + spectrum["binary_len"]

            binary_spectrum_data = binary_data[start:stop]
            n_values_binary = spectrum["n_values"] * float_buffer

            for scan_no in range(spectrum["n_scans"]):
                spec_start = scan_no * n_values_binary
                spec_stop = spec_start + n_values_binary
                scan_data = binary_spectrum_data[spec_start:spec_stop]

                spectrum["data"][f"scan_{scan_no}"] = self._parse_binary_scan(scan_data)

    def _parse_binary_scan(self, binary_scan_data):
        """
        For each spectrum scan, parse the XPS data by unpacking the
        64 bit floats.

        Parameters
        ----------
        binary_spectrum_data : bytes
            Binary data containing the intensity data for one
            spectrum.

        Returns
        -------
        parsed_data : np.ndarray
            One-dimensional array with spectrum intensities.

        """
        float_buffer = self.encoding[1]
        encoding = self.encoding[0]
        parsed_data = []

        n_values = int(len(binary_scan_data) / float_buffer)

        for i in range(n_values):
            start = i * float_buffer
            stop = start + float_buffer
            parsed_data.append(
                struct.unpack_from(encoding, binary_scan_data[start:stop])[0]
            )

        return parsed_data

    def _check_encoding(self):
        """Check if the binary data is single or double encoded."""
        datasize = sum(
            [s["spectrum_header"][8] * s["spectrum_header"][9] for s in self.spectra]
        )
        binary_size = sum([s["spectrum_header"][-2] for s in self.spectra])

        encodings_map = {
            "double": ["<d", 8],
            "float": ["<f", 4],
        }
        if binary_size / datasize == 4:
            self.encoding = encodings_map["float"]
        elif binary_size / datasize == 8:
            self.encoding = encodings_map["double"]
        else:
            logger.error("This binary encoding is not supported.")

    def extract_unit(self, key: str, value):
        """
        Extract units for the metadata containing unit information.

        Example:
            analyser_work_function: 4.506 eV
            -> analyser_work_function: 4.506,
               analyser_work_function_units: eV,

        Parameters
        ----------
        key : str
            Key of the associated value.
        value : str
            Combined unit and value information.

        Returns
        -------
        value :
            value with units.
        unit : str
            Associated unit.

        """
        special_cases = {
            "survey_dwell_time": "s",
            "xray_offset": "um",
            "xray_rotation": "degree",
        }

        try:
            value, unit = split_value_and_unit(value)
        except ValueError:
            unit = ""

        unit = convert_units(unit)

        if key in special_cases:
            unit = special_cases[key]
        if key in UNIT_MISSING:
            unit = UNIT_MISSING[key]

        return value, unit

    def map_values(self, key: str, value, field_type):
        """
        Map values to corresponding structure and field type.

        Parameters
        ----------
        key : str
            PhiMetadata dataclass key.
        value : str
            Original value.
        field_type : class
            Class to be mapped onto.

        Returns
        -------
        value
            Value of correct type and internal structure.

        """
        if key in self.value_function_map:
            map_fn = self.value_function_map[key]
            value = map_fn(value)
        return field_type(value)

    def add_metadata_to_each_spectrum(self):
        """
        Add flattened dict with PhiMetadata fields to each spectrum.

        """
        flattened_metadata = self.flatten_metadata()

        for spectrum in self.spectra:
            spectrum.update(flattened_metadata)
            spectrum["intensity/@units"] = UNIT_MISSING["intensity"]

    def flatten_metadata(self):
        """
        Flatten metadata dict so that key-value pairs of nested
        dictionaries are at the top level.

        Parameters
        ----------
        metadata_dict : dict
            Metadata dict with PhiMetadata fields as keys.

        Returns
        -------
        flattened_dict : dict
            Flatted metadata_dict without any nested substructure.

        """

        def shorten_supkey(supkey: str):
            """Shorted the key for some nested dicts."""
            shortened_key_map = {
                "xray_source": "xray",
                "xray_settings": "xray",
                "stage_positions": "stage",
            }
            if supkey in shortened_key_map:
                return shortened_key_map[supkey]
            return supkey

        flattened_dict = {}

        for key, value in self.metadata.dict().items():
            if isinstance(value, dict):
                for subkey, subvalue in value.items():
                    supkey = shorten_supkey(key)
                    flattened_dict[f"{supkey}_{subkey}"] = subvalue
            else:
                flattened_dict[key] = value

        for key in flattened_dict.copy().keys():
            if "_units" in key:
                new_key = key.replace("_units", "/@units")
                flattened_dict[new_key] = flattened_dict.pop(key)

        for key in flattened_dict.copy().keys():
            if key in KEYS_WITH_UNITS and f"{key}/@units" not in flattened_dict:
                flattened_dict[f"{key}/@units"] = UNIT_MISSING[key]

        return flattened_dict


def _map_file_type(value: str):
    """Map file_type to easily understandable values."""
    value = value.strip()
    file_type_map = {
        "SPECTRUM": "single_spectrum",
        "DEPTHPRO": "depth_profile",
    }

    if value in file_type_map:
        return file_type_map[value]
    return value


def _parse_datetime(value: str):
    """
    Parse datetime into a datetime.datetime object.

    Parameters
    ----------
    value : str
        String representation of the date in the format
        "%m/%d/%y.

    Returns
    -------
    date_object : str
        Datetime in ISO8601 format.

    """
    year, month, day = value.strip().split(" ")
    date_object = datetime.datetime(
        year=int(year), month=int(month), day=int(day), tzinfo=pytz.timezone("UTC")
    )

    return date_object.isoformat()


def _map_to_list(value: str):
    """Map all items in value to a list."""
    try:
        sep = ", " if ", " in value else " "
        values = [float(val) for val in value.split(sep)]
    except ValueError:
        sep = ","
        values = [float(val) for val in value.split(sep)]

    return values


def _map_to_xy(value: str):
    """Map items in value to a dictionary with keys x and y."""
    x, y = value.split(" ")

    return {"x": float(x), "y": float(y)}


def _map_to_xy_with_units(value: str):
    """Map items in value to a dictionary with keys x ."""
    x, y, unit = value.split(" ")
    return {"x": float(x), "x_units": unit, "y": float(y), "y_units": unit}


def _convert_energy_referencing(value: str):
    """Map all items in energy_referencing to a dictionary."""
    peak, energy = value.split(" ")
    return {"peak": peak, "energy": energy, "energy_units": "eV"}


def _convert_channel_info(value: str):
    """Split channel information into list."""
    channel_number, setting_a, setting_b = value.split(" ")
    return [int(setting_a), float(setting_b)]


def _convert_xray_source_params(value: str):
    """Map all items in xray_source_params to a dictionary."""
    label, energy, mono = value.split(" ")

    return {
        "anode_material": label,
        "energy": energy,
        "energy_units": "eV",
        "monochromatized": mono,
    }


def _convert_xray_source_settings(value: str):
    """Map all items in xray_source_settings to a dictionary."""
    (xray_settings) = re.split(r"(\d+)", value)

    for i, setting in enumerate(xray_settings):
        xray_settings[i] = convert_units(setting)

    return {
        "spot_size": float(xray_settings[1]),
        "spot_size_units": xray_settings[2],
        "power": float(xray_settings[3]),
        "power_units": xray_settings[4],
        "high_voltage": float(xray_settings[5]),
        "high_voltage_units": xray_settings[6],
    }


def _convert_stage_positions(value: str):
    """Map all items in stage_positions to a dictionary."""
    x, y, z, azimuth, polar = value.split(" ")

    return {
        "x": float(x),
        "x_units": "mm",
        "y": float(y),
        "y_units": "mm",
        "z": float(z),
        "z_units": "mm",
        "azimuth": float(azimuth),
        "azimuth_units": "degree",
        "polar": float(polar),
        "polar_units": "degree",
    }



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Mappings for Scienta reader.
"""

import re
import datetime
from typing import Dict, Optional
import numpy as np

from pynxtools_xps.reader_utils import (
    convert_pascal_to_snake,
    _re_map_single_value,
)

from pynxtools_xps.value_mappers import (
    convert_energy_type,
    convert_energy_scan_mode,
)


def _extract_energy_units(energy_units: str) -> str:
    """
    Extract energy units from the strings for energy_units.
    Binding Energy [eV] -> eV

    """
    units = re.search(r"\[(.*?)\]", energy_units)
    return units.group(1) if units is not None else "eV"


def _separate_dimension_scale(scale: str) -> np.typing.ArrayLike:
    """
    Seperate the str of the dimension scale into a numpy array

    Parameters
    ----------
    scale : str
        Str of the form "600 599.5 599 598.5 5".

    Returns
    -------
    np.ndarray
        Dimension scale as a numpy array.

    """
    return np.array([float(s) for s in scale.split(" ")])


def _construct_date_time(date_string: str, time_string: str) -> Optional[str]:
    """
    Convert the native time format to the datetime string
    in the ISO 8601 format: '%Y-%b-%dT%H:%M:%S.%fZ'.

    """

    def _parse_date(date_string: str) -> datetime.datetime:
        possible_date_formats = ["%Y-%m-%d", "%m/%d/%Y"]
        for date_fmt in possible_date_formats:
            try:
                return datetime.datetime.strptime(date_string, date_fmt)
            except ValueError as err:
                continue
        raise ValueError("Date format not recognized")

    def _parse_time(time_string: str) -> datetime.time:
        possible_time_formats = ["%H:%M:%S", "%I:%M:%S %p"]
        for time_fmt in possible_time_formats:
            try:
                return datetime.datetime.strptime(time_string, time_fmt).time()
            except ValueError as err:
                continue
        raise ValueError("Time format not recognized")

    try:
        date = _parse_date(date_string)
        time = _parse_time(time_string)
        datetime_obj = datetime.datetime.combine(date, time)

        return datetime_obj.astimezone().isoformat()

    except ValueError as err:
        raise ValueError(
            "Date and time could not be converted to ISO 8601 format."
        ) from err


KEY_MAP: Dict[str, str] = {
    "number_of_regions": "no_of_regions",
    "version": "software_version",
    "dimension_name": "energy_units",
    "dimension_size": "energy_size",
    "dimension_scale": "energy_axis",
    "number_of_sweeps": "no_of_scans",
    "energy_unit": "energy_scale_2",
    "low_energy": "start_energy",
    "high_energy": "stop_energy",
    "energy_step": "step_size",
    "step_time": "dwell_time",
    "detector_first_x-_channel": "detector_first_x_channel",
    "detector_last_x-_channel": "detector_last_x_channel",
    "detector_first_y-_channel": "detector_first_y_channel",
    "detector_last_y-_channel": "detector_last_y_channel",
    "file": "data_file",
    "sequence": "sequence_file",
    "spectrum_name": "spectrum_type",
    "instrument": "instrument_name",
    "location": "vendor",
    "user": "user_name",
    "sample": "sample_name",
    "comments": "spectrum_comment",
    "date": "start_date",
    "time": "start_time",
}

VALUE_MAP = {
    "no_of_regions": int,
    "energy_size": int,
    "pass_energy": float,
    "no_of_scans": int,
    "excitation_energy": float,
    "center_energy": float,
    "start_energy": float,
    "stop_energy": float,
    "step_size": float,
    "dwell_time": float,
    "detector_first_x_channel": int,
    "detector_last_x_channel": int,
    "detector_first_y_channel": int,
    "detector_last_y_channel": int,
    "time_per_spectrum_channel": float,
    "energy_units": _extract_energy_units,
    "energy_axis": _separate_dimension_scale,
    "energy_scale": convert_energy_type,
    "energy_scale_2": convert_energy_type,
    "acquisition_mode": convert_energy_scan_mode,
    "time_per_spectrum_channel": float,
    "manipulator_r1": float,
    "manipulator_r2": float,
}

UNITS: dict = {
    "energydispersion/pass_energy": "eV",
    "beam_xray/excitation_energy": "eV",
    "region/energy_axis": "eV",
    "region/center_energy": "eV",
    "region/start_energy": "eV",
    "region/stop_energy": "eV",
    "region/step_size": "eV",
    "detector/dwell_time": "eV",
    "region/time_per_spectrum_channel": "s",
}


def _get_key_value_pair(line: str):
    """
    Split the line at the '=' sign and return a
    key-value pair. The values are mapped according
    to the desired format.

    Parameters
    ----------
    line : str
        One line from the input file.

    Returns
    -------
    Tuple[str, object]
        A tuple containing:
        - key : str
            Anything before the '=' sign, mapped to the desired
            key format.
        - value : object
            Anything after the '=' sign, mapped to the desired
            value format and type.

    """
    try:
        key, value = line.split("=")
        key = convert_pascal_to_snake(key)
        key = KEY_MAP.get(key, key)
        if "dimension" in key:
            key_part = f"dimension_{key.rsplit('_')[-1]}"
            key = KEY_MAP.get(key_part, key_part)
        value = _re_map_single_value(key, value, VALUE_MAP)

    except ValueError:
        key, value = "", ""

    return key, value



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-lines,too-few-public-methods
"""
A generic reader for loading XPS (X-ray Photoelectron Spectroscopy) data
file into mpes nxdl (NeXus Definition Language) template.
"""

import copy
import datetime
import logging
import os
import re
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import numpy as np
from pynxtools.dataconverter.helpers import extract_atom_types
from pynxtools.dataconverter.readers.multi.reader import MultiFormatReader
from pynxtools.dataconverter.readers.utils import parse_yml
from pynxtools.dataconverter.template import Template

from pynxtools_xps.reader_utils import check_units

from pynxtools_xps.phi.spe_pro_phi import MapperPhi
from pynxtools_xps.scienta.scienta_reader import MapperScienta
from pynxtools_xps.specs.sle.sle_specs import SleMapperSpecs
from pynxtools_xps.specs.xml.xml_specs import XmlMapperSpecs
from pynxtools_xps.specs.xy.xy_specs import XyMapperSpecs
from pynxtools_xps.vms.vamas_export import TxtMapperVamasExport, CsvMapperVamasResult
from pynxtools_xps.vms.vamas import VamasMapper

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


np.set_printoptions(threshold=sys.maxsize)


CONVERT_DICT = {
    "unit": "@units",
    "version": "@version",
    "user": "USER[user]",
    "instrument": "INSTRUMENT[instrument]",
    "source_xray": "sourceTYPE[source_xray]",
    "beam_xray": "beamTYPE[beam_xray]",
    "analyser": "ELECTRONANALYSER[electronanalyser]",
    "collectioncolumn": "COLLECTIONCOLUMN[collectioncolumn]",
    "energydispersion": "ENERGYDISPERSION[energydispersion]",
    "detector": "DETECTOR[detector]",
    "manipulator": "MANIPULATOR[manipulator]",
    "pid": "PID[pid]",
    "process": "PROCESS[process]",
    "sample": "SAMPLE[sample]",
    "substance": "SUBSTANCE[substance]",
}

REPLACE_NESTED: Dict[str, str] = {}

CHAN_COUNT = "_chan"
SCAN_COUNT = "_scan"


def _get_channel_vars(data_vars: List[str]):
    """Get all data vars that containt _chan."""
    return [data_var for data_var in data_vars if CHAN_COUNT in data_var]


def _get_scan_vars(data_vars: List[str]):
    """Get all data vars that contain _scan, but not _chan."""
    return [
        data_var
        for data_var in data_vars
        if SCAN_COUNT in data_var and CHAN_COUNT not in data_var
    ]


def concatenate_values(value1, value2):
    """
    Concatenate two values of same type to be stored
    in xps_data_dict. Dicts are merged and every other object is
    appended to a list.

    """
    if isinstance(value1, dict) and isinstance(value2, dict):
        concatenated = {**value1, **value2}
    else:
        if not isinstance(value1, list):
            value1 = [value1]
        if not isinstance(value2, list):
            value2 = [value2]
        concatenated = value1 + value2

    return concatenated


def _check_multiple_extensions(file_paths: Tuple[str] = None) -> bool:
    """
    Determines if a list of file paths contains more than one unique file extension.

    This method accepts a list of file paths (as strings or `Path` objects) and checks
    if there are multiple unique file extensions present in the list. A file extension
    is identified as the substring after the last period (`.`) in the file name.

    Parameters:
        file_paths (Tuple[str]): A tuple of file paths, which can be strings or
                                 `Path` objects. Defaults to None.

    Returns:
        bool: True if more than one unique file extension is found, False otherwise.

    Raises:
        TypeError: If `file_paths` is not a tuple of strings or `Path` objects.
    """
    extensions = {str(path).split(".")[-1] for path in file_paths if "." in str(path)}

    return len(extensions) > 1


# pylint: disable=too-few-public-methods
class XPSReader(MultiFormatReader):
    """Reader for XPS."""

    supported_nxdls = [
        "NXmpes",
        "NXxps",
    ]

    reader_dir = Path(__file__).parent
    config_file: Optional[Union[str, Path]] = reader_dir.joinpath(
        "config", "template.json"
    )

    __prmt_file_ext__ = [
        ".ibw",
        ".npl",
        ".pro",
        ".spe",
        ".sle",
        ".slh",
        ".vms",
        ".xml",
        ".xy",
        ".txt",  # This is last because of the processing_order
    ]

    __prmt_metadata_file_ext__ = {".csv": ".txt"}

    __vendors__ = ["kratos", "phi", "scienta", "specs", "unkwown"]
    __prmt_vndr_cls: Dict[str, Dict] = {
        ".csv": {"unknown": CsvMapperVamasResult},
        ".ibw": {"scienta": MapperScienta},
        ".npl": {"unkwown": VamasMapper},
        ".pro": {"phi": MapperPhi},
        ".spe": {"phi": MapperPhi},
        ".sle": {"specs": SleMapperSpecs},
        ".txt": {
            "scienta": MapperScienta,
            "unknown": TxtMapperVamasExport,
        },
        ".vms": {"unkwown": VamasMapper},
        ".xml": {"specs": XmlMapperSpecs},
        ".xy": {"specs": XyMapperSpecs},
    }

    __file_err_msg__ = (
        "Need an XPS data file with one of the following extensions: "
        f"data files: {__prmt_file_ext__}, metadata files: {__prmt_metadata_file_ext__}."
    )

    __vndr_err_msg__ = (
        f"Need an XPS data file from one of the following vendors: {__vendors__}"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.xps_data_dicts: List[Dict[str, Any]] = []
        self.xps_data: Dict[str, Any] = {}
        self.eln_data: Dict[str, Any] = {}

        self.extensions = {
            ".yml": self.handle_eln_file,
            ".yaml": self.handle_eln_file,
            ".json": self.set_config_file,
        }

        self.processing_order = (
            XPSReader.__prmt_file_ext__
            + list(XPSReader.__prmt_metadata_file_ext__.keys())
            + list(self.extensions.keys())
        )

        for ext in XPSReader.__prmt_file_ext__ + list(
            XPSReader.__prmt_metadata_file_ext__.keys()
        ):
            self.extensions[ext] = self.handle_data_file

        self.processing_order = XPSReader.__prmt_file_ext__ + [
            ".yml",
            ".yaml",
            ".json",
        ]

    def set_config_file(self, file_path: str) -> Dict[str, Any]:
        if self.config_file is not None:
            logger.info(
                f"Config file already set. Replaced by the new file {file_path}."
            )
        self.config_file = file_path
        return {}

    def handle_eln_file(self, file_path: str) -> Dict[str, Any]:
        """
        Loads ELN file and handles specific cases.
        """

        def combine_and_unique_string(string: str, elements: List[str]) -> str:
            """
            Combines a comma-separated string and a list into a single string with unique elements.

            Args:
                string (str): A comma-separated string.
                elements (list): A list of elements to combine with the string.

            Returns:
                str: A comma-separated string with unique elements.
            """
            existing_elements = [
                item.strip() for item in string.split(",") if item.strip()
            ]
            combined_elements = list(set(existing_elements + elements))
            combined_elements.sort()
            return ", ".join(combined_elements)

        eln_data = parse_yml(
            file_path,
            convert_dict=CONVERT_DICT,
            replace_nested=REPLACE_NESTED,
            parent_key="/ENTRY",
        )

        # replace paths for entry-specific ELN data
        pattern = re.compile(r"(/ENTRY)/ENTRY(\[[^\]]+\])")

        formula_keys = ("molecular_formula_hill", "chemical_formula")

        initial_eln_keys = list(eln_data.keys())

        for key, value in eln_data.copy().items():
            new_key = pattern.sub(r"\1\2", key)

            # Parse substance/molecular_formula_hill and chemical_formula into atom_types
            for form_key in formula_keys:
                if form_key in key:
                    atom_types = list(extract_atom_types(value))

                    if atom_types:
                        modified_key = re.sub(r"SUBSTANCE\[.*?\]/", "", key)
                        modified_key = modified_key.replace(form_key, "atom_types")

                        if modified_key not in initial_eln_keys:
                            if modified_key not in self.eln_data:
                                self.eln_data[modified_key] = ", ".join(atom_types)
                            else:
                                self.eln_data[modified_key] = combine_and_unique_string(
                                    self.eln_data[modified_key], atom_types
                                )
                        else:
                            logger.info(
                                f"{key} from ELN was not parsed to atom_types because {modified_key} already exists."
                            )

            if isinstance(value, datetime.datetime):
                eln_data[key] = value.isoformat()

            self.eln_data[new_key] = eln_data.pop(key)

        return {}

    def handle_data_file(self, file_path: str) -> Dict[str, Any]:
        def _check_for_vendors(file_path: str) -> str:
            """
            Check for the vendor name of the XPS data file.

            """
            _, file_ext = os.path.splitext(file_path)

            vendor_dict = XPSReader.__prmt_vndr_cls[file_ext]

            if len(vendor_dict) == 1:
                return list(vendor_dict.keys())[0]
            if file_ext == ".txt":
                return _check_for_vendors_txt(file_path)
            raise ValueError(XPSReader.__vndr_err_msg__)

        def _check_for_vendors_txt(file_path: str) -> str:
            """
            Search for a vendor names in a txt file

            Parameters
            ----------
            file : str
                XPS txt file.

            Returns
            -------
            vendor
                Vendor name if that name is in the txt file.

            """
            vendor_dict = XPSReader.__prmt_vndr_cls[".txt"]

            with open(file_path, encoding="utf-8") as txt_file:
                contents = txt_file.read()

            for vendor in vendor_dict:
                vendor_options = [vendor, vendor.upper(), vendor.capitalize()]

                if any(vendor_opt in contents for vendor_opt in vendor_options):
                    return vendor
                if contents[:6] == "[Info]":
                    # This is for picking the Scienta reader if "scienta"
                    # is not in the file
                    return vendor
            return "unknown"

        _, file_ext = os.path.splitext(file_path)

        if file_ext in XPSReader.__prmt_file_ext__:
            vendor = _check_for_vendors(file_path)

            parser = XPSReader.__prmt_vndr_cls[file_ext][vendor]()
            parser.parse_file(file_path, **self.kwargs)
            data_dict = parser.data_dict

            self.config_file = XPSReader.reader_dir.joinpath(
                "config", parser.config_file
            )
            self.xps_data_dicts += [data_dict]

        elif file_ext in XPSReader.__prmt_metadata_file_ext__:
            vendor = _check_for_vendors(file_path)

            metadata_parser = XPSReader.__prmt_vndr_cls[file_ext][vendor]()
            metadata_parser.parse_file(file_path, **self.kwargs)

            main_file_ext = XPSReader.__prmt_metadata_file_ext__[file_ext]

            main_file_dicts = [
                d for d in self.xps_data_dicts if d.get("file_ext") == main_file_ext
            ]

            metadata_parser.update_main_file_dict(main_file_dicts)

        return {}

    def get_entry_names(self) -> List[str]:
        """
        Returns a list of entry names which should be constructed from the data.
        Defaults to creating a single entry named "entry".
        """
        # Track entries for using for eln data
        entries: List[str] = []

        try:
            for entry in self.xps_data["data"]:
                entries += [entry]
        except KeyError:
            pass

        if not entries:
            entries += ["entry"]

        return list(dict.fromkeys(entries))

    def setup_template(self) -> Dict[str, Any]:
        """
        Setups the initial data in the template.
        """
        # TODO: Set fixed information, e.g., about the reader.
        return {}

    def handle_objects(self, objects: Tuple[Any]) -> Dict[str, Any]:
        """
        Handles the objects passed into the reader.
        """
        return {}

    def post_process(self) -> None:
        """
        Do postprocessing after all files and the config file are read .
        """
        self._combine_datasets()

        # TODO: make processing of multiple entities robust
        # self.process_multiple_entities()

    def _combine_datasets(self) -> None:
        """
        This function (which as called after all files have been read)
        combines the different data sets from the XPS files and ensures
        that entry names are different and that no data is overwritten.

        Returns
        -------
        None

        """

        def check_for_same_entries(
            dicts: List[Dict[str, Any]],
        ) -> Optional[Tuple[Set[str], List[Set[int]]]]:
            """
            Checks whether the input dictionaries have the same entries and identifies which
            dictionaries contain each common entry.

            Args:
                dicts (List[Dict[str, any]]): A list of dictionaries with keys potentially containing entries.

            Returns:
                Optional[Tuple[Set[str], List[Set[int]]]]:
                - A tuple where:
                  - The first element is a set of common entries found across the dictionaries.
                  - The second element is a list of sets, where each set contains the indices of dictionaries
                    that have the corresponding entry.
                - None if no common entries are found.
            """
            entry_pattern = re.compile(r"/ENTRY\[(.*?)\]")

            entry_to_dicts: Dict[str, Set] = {}

            for i, d in enumerate(dicts):
                for key in d.keys():
                    entries = entry_pattern.findall(key)
                    for entry in entries:
                        if entry not in entry_to_dicts:
                            entry_to_dicts[entry] = set()
                        entry_to_dicts[entry].add(i)

            common_entries = {
                entry
                for entry, dict_indices in entry_to_dicts.items()
                if len(dict_indices) > 1
            }

            if not common_entries:
                return None, None

            dict_indices = [entry_to_dicts[entry] for entry in common_entries]

            return common_entries, dict_indices

        common_entries, dict_indices = check_for_same_entries(self.xps_data_dicts)

        if common_entries and not self.overwrite_keys:
            for entry, indices in zip(common_entries, dict_indices):
                dicts_with_common_entries = [self.xps_data_dicts[i] for i in indices]

                for i, data_dict in enumerate(dicts_with_common_entries):
                    for key, value in data_dict.copy().items():
                        new_key = key.replace(f"/ENTRY[{entry}]", f"/ENTRY[{entry}{i}]")
                        if key == "data":
                            for entry_name, xarr in value.copy().items():
                                if entry_name == entry:
                                    new_entry_name = entry_name.replace(
                                        f"{entry}", f"{entry}{i}"
                                    )
                                    value[new_entry_name] = xarr
                                    del value[entry_name]
                        if new_key != key:
                            data_dict[new_key] = value
                            del data_dict[key]

        for data_dict in self.xps_data_dicts:
            # If there are multiple input data files of the same type,
            # make sure that existing keys are not overwritten.
            existing = [
                (key, self.xps_data[key], data_dict[key])
                for key in set(self.xps_data).intersection(data_dict)
            ]

            self.xps_data = {**self.xps_data, **data_dict}

            if not self.overwrite_keys:
                for key, value1, value2 in existing:
                    self.xps_data[key] = concatenate_values(value1, value2)

    def _get_analyser_names(self) -> List[str]:
        """
        Returns a list of analyser names which should be constructed
        from the data. Defaults to creating a single analyser named
        "analyser".

        Currently, this is not used, but can be changed if there are
        multiple analysers in the future.
        """
        analysers: List[str] = []

        if not analysers:
            analysers += ["electronanalyser"]

        return list(dict.fromkeys(analysers))

    def _get_detector_names(self) -> List[str]:
        """
        Returns a list of detector names which should be constructed
        from the data. Defaults to creating a single detector named
        "detector".
        """
        detectors: List[str] = []

        try:
            for entry, entry_values in self.xps_data["data"].items():
                for data_var in entry_values:
                    if CHAN_COUNT in data_var:
                        detector_num = data_var.split(CHAN_COUNT)[-1]
                        detector_nm = f"detector{detector_num}"
                        detectors += [detector_nm]
        except KeyError:
            pass

        if not detectors:
            detectors += ["detector"]

        return list(dict.fromkeys(detectors))

    def process_multiple_entities(self) -> None:
        """
        Check if there are multiple of some class and, if so, change the
        keys and values in the config file.

        This replaces all occureces of "detector" and "electronanalyser"
        in the config dict by the respective names (e.g., detector0, detector1)
        and removes the generic term if there are multiple different instances.

        """
        multiples_to_check = {
            "electronanalyser": self._get_analyser_names,
            "detector": self._get_detector_names,
        }

        """
        Currently, it only works if the config_dict is loaded BEFORE the
        parse_json_config method".

        In principle, the same replacement should be done for the eln and
        (meta)data dicts.
        """

        for config_key, config_value in self.config_dict.copy().items():
            for original_key, search_func in multiples_to_check.items():
                entity_names = search_func()

                if len(entity_names) >= 1 and entity_names[0] is not original_key:
                    for name in entity_names:
                        modified_value = copy.deepcopy(config_value)

                        modified_key = config_key.replace(
                            f"[{original_key}]",
                            f"[{name}]",
                        )

                        if (
                            isinstance(config_value, str)
                            and f"{original_key}/" in config_value
                        ):
                            modified_value = config_value.replace(
                                f"{original_key}/", f"{name}/"
                            )

                        self.config_dict[modified_key] = modified_value
                        del self.config_dict[config_key]

    def get_metadata(
        self,
        metadata_dict: Dict[str, Any],
        path: str,
        entry_name: str,
    ) -> Any:
        """
        Get metadata from the ELN or XPS data dictionaries.

        Note that the keys of metadata_dict may contain more than
        the path, i.e.,
        /ENTRY[my-entry]/instrument/analyser/collectioncolumn/voltage.
        With the regex, the path = "collectioncolumn/voltage" would
        still yield the correct value.

        Parameters
        ----------
        metadata_dict : Dict[str, Any]
            One of ELN or XPS data dictionaries .
        path : str
            Path to search in the metadata_dict
        entry_name : str
            Entry name to search.

        Yields
        ------
        value: Any
            The value in the metadata_dict.

        """
        pattern = re.compile(
            rf"^/ENTRY\[{re.escape(entry_name)}\](?:/.*/|/){re.escape(path)}$"
        )

        matching_key = next((key for key in metadata_dict if pattern.match(key)), None)

        value = metadata_dict.get(matching_key)

        if value is None or str(value) == "None":
            return

        if isinstance(value, datetime.datetime):
            value = value.isoformat()

        return value

    def get_attr(self, key: str, path: str) -> Any:
        """
        Get the metadata that was stored in the main file.
        """
        return self.get_metadata(self.xps_data, path, self.callbacks.entry_name)

    def get_eln_data(self, key: str, path: str) -> Any:
        """
        Returns data from the given eln path.
        Gives preference to ELN data for a given entry before searching
        the ELN data for all entries.
        Returns None if the path does not exist.
        """
        if key in self.eln_data:
            return self.eln_data.get(key)

        else:
            # check for similar key with generic /ENTRY/
            pattern = re.compile(r"(/ENTRY)\[[^\]]+\]")
            modified_key = pattern.sub(r"\1", key)
            if modified_key in self.eln_data:
                return self.eln_data.get(modified_key)
        return

    def get_data_dims(self, key: str, path: str) -> List[str]:
        """
        Returns the dimensions of the data from the given path.
        """

        def get_signals(key: str) -> List[str]:
            xr_data = self.xps_data["data"].get(f"{self.callbacks.entry_name}")

            if key == "scans":
                data_vars = _get_scan_vars(xr_data.data_vars)
            elif key == "channels":
                data_vars = _get_channel_vars(xr_data.data_vars)
                if not data_vars:
                    data_vars = _get_scan_vars(xr_data.data_vars)
            else:
                data_vars = [""]

            return list(map(str, data_vars))

        def get_processes(process_key: str) -> List[str]:
            escaped_entry_name = re.escape(self.callbacks.entry_name)

            pattern = re.compile(
                rf"/ENTRY\[{escaped_entry_name}]\.*/{process_key}([a-zA-Z0-9_]+)"
            )

            process_names = {
                match for key in self.xps_data for match in pattern.findall(key)
            }

            return sorted(process_names)

        patterns: Dict[str, Any] = {
            "peak": lambda: get_processes("component"),
            "background": lambda: get_processes("region"),
            r"data/DATA": lambda: get_signals(path.split(":*.")[-1]),
            r"DETECTOR\[[a-zA-Z0-9_]+\]/raw_data": lambda: get_signals("channels"),
        }

        for pattern, func in patterns.items():
            if re.search(pattern, key):
                return func()

        return get_signals(key="scans")

    def get_data(self, key: str, path: str) -> Any:
        """
        Returns data for a given key.
        Can either return averaged, scan, or channel data.
        Should return None if the path does not exist.
        """
        xr_data = self.xps_data["data"].get(f"{self.callbacks.entry_name}")

        if path.endswith("average"):
            return np.mean(
                [xr_data[x_arr].data for x_arr in _get_scan_vars(xr_data.data_vars)],
                axis=0,
            )

        elif path.endswith("errors"):
            return np.std(
                [xr_data[x_arr].data for x_arr in _get_scan_vars(xr_data.data_vars)],
                axis=0,
            )

        elif path.endswith("raw_data"):
            data_vars = _get_channel_vars(xr_data.data_vars)

            if not data_vars:
                # If there is no channel data, use scan data.
                data_vars = _get_scan_vars(xr_data.data_vars)

            # Skip average cycle data
            return np.array(
                [
                    xr_data[data_var].data
                    for data_var in data_vars
                    if SCAN_COUNT in data_var
                ]
            )

        elif path.endswith("scans"):
            return np.array(xr_data[path.split(".scans")[0]])

        elif path.endswith("channels"):
            return np.array(xr_data[path.split(".channels")[0]])

        elif "energy" in path:
            return np.array(xr_data.coords["energy"].values)

        else:
            try:
                return xr_data[path]
            except KeyError:
                try:
                    return np.array(xr_data.coords[path].values)
                except KeyError:
                    pass

    def set_nxdata_defaults(self, template):
        """Set the default for automatic plotting."""
        survey_count, count = 0, 0

        def get_unique_nxfit_names(template) -> Set[str]:
            """Extract unique 'ENTRY[<some-name>]/FIT[<some-other-name>]' pairs from template keys."""
            pattern = re.compile(r"^/?ENTRY\[(?P<entry>[^]]+)\]/FIT\[(?P<fit>[^]]+)\]/")

            result = set()
            for key in template:
                m = pattern.match(key)
                if m:
                    result.add(f"{m.group('entry')}/{m.group('fit')}")
            return result

        def get_first_matching_fit(
            entry_name: str, unique_fits: Set[str]
        ) -> Optional[str]:
            """Return the first '<fit>' name that matches the given entry name, if any."""
            for fit in unique_fits:
                if fit.startswith(f"{entry_name}/"):
                    return fit.split("/", 1)[1]  # Extract only the fit name
            return None

        unique_fits = sorted(
            get_unique_nxfit_names(template)
        )  # Sorting for deterministic ordering

        for entry in self.get_entry_names():
            if unique_fits:
                template["/@default"] = unique_fits[0]
                match = get_first_matching_fit(entry, unique_fits)
                if match:
                    template[f"/ENTRY[{entry}]/@default"] = match

                else:
                    template[f"/ENTRY[{entry}]/@default"] = "data"

            else:
                if "Survey" in entry and survey_count == 0:
                    survey_count += 1
                    template["/@default"] = entry

                # If no Survey, set any scan for default
                elif survey_count == 0 and count == 0:
                    count += 1
                    template["/@default"] = entry

    def read(
        self,
        template: dict = None,
        file_paths: Tuple[str] = None,
        objects: Tuple[Any] = None,
        **kwargs,
    ) -> dict:
        self.overwrite_keys = _check_multiple_extensions(file_paths)

        template = super().read(template, file_paths, objects, suppress_warning=True)
        self.set_nxdata_defaults(template)

        final_template = Template()
        for key, val in template.items():
            if key.endswith("@units"):
                parent = key.replace("/@units", "")
                if parent not in template:
                    continue
            if val is not None:
                if "@units" in key:
                    check_units(key, val)
                final_template[key] = val

        return final_template


READER = XPSReader



"""Fitting functions for XPS spectra"""

from typing import Optional
from dataclasses import dataclass
import numpy as np
from numpy.linalg import norm
import plotly.graph_objects as go
import pandas as pd
from h5py import File as H5File
from lmfit import Model, CompositeModel


@dataclass
class XPSRegion:
    """An XPS region representation"""

    binding_energy: np.ndarray
    counts: np.ndarray
    counts_err: np.ndarray
    baseline: Optional[np.ndarray] = None
    _fit_region: slice = slice(None, None)
    _fit_mod: Optional[Model] = None
    fit_result: Optional[Model] = None

    @staticmethod
    def load(filename: str, entry: str) -> "XPSRegion":
        """Load from a NeXus file.

        Args:
            filename (str): The NeXus file name to load.
            entry (str):
                The entry from which to load data.
                Should be the name of an NXentry within the NeXus file.

        Returns:
            XPSRegion: The XPSRegion class with the loaded data.
        """
        with H5File(filename, "r") as xps_file:
            binding_energy = xps_file[f"/{entry}/data/energy"][:]
            cps = xps_file[f"/{entry}/data/data"][:]
            cps_err = xps_file[f"/{entry}/data/data_errors"][:]

        return XPSRegion(binding_energy=binding_energy, counts=cps, counts_err=cps_err)

    def fit_region(self, start: int, stop: int) -> "XPSRegion":
        """Select a fit region within this XPSregion by x-axis value.
        The fit region is always selected between start and stop, regardless of their order.
        Both points are included in the region,
        hence the actual selected region may be a little larger.

        Args:
            start (int): The start ot the region.
            stop (int): The end of the region.

        Returns:
            XPSRegion: This class
        """
        region = np.argwhere(
            (self.binding_energy >= start) & (self.binding_energy <= stop)
        )

        self._fit_region = slice(region[0, 0], region[-1, 0], 1)
        return self

    def fit_model(self, model: Model) -> "XPSRegion":
        """Supply a fit model to fit this xps region.

        Args:
            model (lmfit.Model): The lmfit model to use.

        Returns:
            XPSRegion: This class
        """
        self._fit_mod = model
        return self

    def fit(self, *args, **kwargs) -> "XPSRegion":
        """Perform a fit of the data. You need to define a fit_model first and
        execute a baseline correction before using this method.

        Raises:
            ValueError: If no fit model is provided or the baseline has not been corrected.

        Returns:
            XPSRegion: This class
        """
        if self._fit_mod is None:
            raise ValueError("You need to provide a fit model before performing a fit.")

        if self.baseline is None:
            raise ValueError(
                "You need to perform a baseline correction before using this method."
            )

        self.fit_result = self._fit_mod.fit(
            (
                self.counts[self._fit_region] - self.baseline
                if self._fit_region
                else self.counts[self._fit_region]
            ),
            *args,
            x=self.binding_energy[self._fit_region],
            weights=1 / self.counts_err[self._fit_region],
            **kwargs,
        )

        return self

    def calc_baseline(self, bg_type: str = "shirley") -> "XPSRegion":
        """Calculate the baseline for this xps spectrum in the given region.

        Args:
            bg_type (str, optional): The background type. Defaults to "shirley".

        Raises:
            ValueError: If the bg_type is unsupported.

        Returns:
            XPSRegion: This class
        """
        baselines = {"shirley": shirley_baseline}
        if bg_type not in baselines:
            raise ValueError(f"Unsupported baseline type {bg_type}.")

        self.baseline = baselines[bg_type](
            self.binding_energy[self._fit_region], self.counts[self._fit_region]
        )

        return self

    def peak_property(self, prop: str) -> pd.DataFrame:
        """Generates a dataframe with values for a property `prop` of a fitting model.

        Args:
            prop (str):
                The name of the property to deduce for the peaks,
                e.g. `center` for the peak center for gaussian or lorentzian shapes.

        Raises:
            ValueError:  Thrown if no prior fit is performed.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the peak property for each peak.
        """
        if not self.fit_result:
            raise ValueError("You need to perform a fit first.")

        props = pd.DataFrame()
        for prefix in map(lambda x: x.prefix, self.fit_result.components):
            if f"{prefix}{prop}" not in self.fit_result.params:
                continue

            props = pd.concat(
                [
                    props,
                    pd.DataFrame(
                        {prop: self.fit_result.params.get(f"{prefix}{prop}").value},
                        index=[prefix.rstrip("_")],
                    ),
                ]
            )

        return props

    def peak_areas(self, region_only=False) -> pd.DataFrame:
        """Calculates the peak areas of the given fit models peaks.

        Args:
            region_only (bool, optional):
                Set true if only the area inside the set region should be consider.
                Defaults to False.

        Raises:
            ValueError: Thrown if no prior fit is performed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the peak areas.
        """
        if not self.fit_result:
            raise ValueError("You need to perform a fit first.")

        areas = pd.DataFrame()
        if region_only:
            peaks = self.fit_result.eval_components(
                x=self.binding_energy[self._fit_region]
            )
        else:
            peaks = self.fit_result.eval_components(x=self.binding_energy)
        for prefix in peaks:
            areas = pd.concat(
                [
                    areas,
                    pd.DataFrame(
                        {"Area": sum(peaks[prefix])},
                        index=[prefix.rstrip("_")],
                    ),
                ]
            )
        return areas

    def plot_residual(self):
        """Plot the fit residual"""
        if not self.fit_result:
            raise ValueError("You need to perform a fit first.")

        fig = go.Figure(
            data=go.Scatter(
                name="Residual",
                x=self.binding_energy[self._fit_region],
                y=self.fit_result.residual,
            )
        )
        fig.update_xaxes(title="Binding energy (eV)")
        fig.update_yaxes(title="Residual")
        fig.layout.xaxis.autorange = "reversed"

        fig.show()

    def plot(self):
        """Plot the xps region"""
        fig = go.Figure(
            data=go.Scatter(
                name="Measurement",
                x=self.binding_energy,
                y=self.counts,
                error_y=dict(
                    type="data",  # value of error bar given in data coordinates
                    array=self.counts_err,
                    visible=True,
                ),
            )
        )
        if self._fit_region.start is not None:
            fig.add_vline(
                self.binding_energy[self._fit_region.start], line_color="lightgrey"
            )
        if self._fit_region.stop is not None:
            fig.add_vline(
                self.binding_energy[self._fit_region.stop], line_color="lightgrey"
            )
        if self.baseline is not None:
            fig.add_trace(
                go.Scatter(
                    name="Baseline",
                    x=self.binding_energy[self._fit_region],
                    y=self.baseline,
                ),
            )
        if self.fit_result is not None:
            fig.add_trace(
                go.Scatter(
                    name="Fit",
                    x=self.binding_energy[self._fit_region],
                    y=self.fit_result.best_fit + self.baseline,
                ),
            )
        if isinstance(self._fit_mod, CompositeModel):
            peaks = self.fit_result.eval_components(
                x=self.binding_energy[self._fit_region]
            )
            for prefix in peaks:
                fig.add_trace(
                    go.Scatter(
                        name=prefix.rstrip("_"),
                        x=self.binding_energy[self._fit_region],
                        y=peaks[prefix] + self.baseline,
                    )
                )

        fig.update_xaxes(title="Binding energy (eV)")
        fig.update_yaxes(title="CPS")
        fig.layout.xaxis.autorange = "reversed"

        fig.show()


# pylint: disable=invalid-name
def shirley_baseline(
    x: np.ndarray, y: np.ndarray, tol: float = 1e-5, maxit: int = 10
) -> np.ndarray:
    """Calculate the shirley background according to the Sherwood method.

    Args:
        x (np.ndarray): The x-axis on which to calculate the shirley baseline.
        y (np.ndarray): The y-axis on which to calculate the shirley baseline.
        tol (float, optional):
            The convergence tolerance at which to stop the iteration.
            Defaults to 1e-5.
        maxit (int, optional):
            The maximum iteration after which to stop the iteration.
            Defaults to 10.

    Raises:
        ValueError:
            Is thrown when the arrays have different dimensions,
            are not numpy arrays or are empty.
            Is also thrown when the fit does not converge.

    Returns:
        np.ndarray: The shirley baseline for the x, y dataset.
    """

    if not isinstance(x, np.ndarray) or not isinstance(y, np.ndarray):
        raise ValueError(
            f"Parameters x and y must be of type numpy array, not {type(x)} and {type(y)}"
        )

    if len(x) != len(y):
        raise ValueError("x and y arrays have different dimensions.")

    if not x.any():
        raise ValueError("x-array is empty.")

    if len(x.shape) > 1:
        raise ValueError(
            f"Data arrays must be one-dimensional. Found dimension {x.shape}."
        )

    is_reversed = False
    if x[0] < x[-1]:
        is_reversed = True
        x = x[::-1]
        y = y[::-1]

    background = np.zeros(x.shape)
    background_next = background.copy()

    iters = 0
    while True:
        k = (y[0] - y[-1]) / np.trapz(y - background, x=x)

        for energy in range(len(x)):
            background_next[energy] = k * np.trapz(
                y[energy:] - background[energy:], x=x[energy:]
            )

        diff = norm(background_next - background)
        background = background_next.copy()
        if diff < tol:
            break

        iters += 1
        if iters == maxit:
            raise ValueError(
                "Maximum number of iterations exceeded before convergence."
            )

    if is_reversed:
        return (y[-1] + background)[::-1]
    return y[-1] + background
